# -*- coding: utf-8 -*-
"""Classes and functions to evaluate embeddings according to the "Outlier Detection" task.


This module implements the evaluation task defined in [1]_

Datasets available in this module :

* `OD_8_8_8` [1]_
* `WIKI_SEM_500` [2]_

References
----------
.. [1] José Camacho-Collados and Roberto Navigli. Find the word that does not belong: A Framework for an Intrinsic
        Evaluation of Word Vector Representations. In Proceedings of the ACL Workshop on Evaluating Vector Space
        Representations for NLP, Berlin, Germany, August 12, 2016.
.. [2]


"""
from collections import namedtuple

import numpy as np

import mangoes
import mangoes.utils.arrays
import mangoes.utils.metrics
from mangoes.evaluation.base import PrintableReport, BaseEvaluation, BaseEvaluator


# ####################################################################################################################
# DATASETS

class Dataset(mangoes.evaluation.base.BaseDataset):
    """Class to create a Dataset for outlier detection task, to be used in Evaluation class

    The outlier is the last word of the group

    Examples
    --------
    >>> from mangoes.evaluation.outlier import Dataset
    >>> user_dataset = Dataset("user dataset", ['january february march saturn', 'monday tuesday friday phone'])
    >>> cats_dataset = Dataset("cats", "../resources/en/outlier_detection/8-8-8/Big_cats.txt")

    2 analogy datasets are available in this module:

    - the 8-8-8 dataset :
    >>> import mangoes.evaluation.outlier
    >>> _8_8_8 = mangoes.evaluation.outlier._8_8_8

    - the Wiki Sem 500 dataset :
    >>> import mangoes.evaluation.outlier
    >>> msr = mangoes.evaluation.outlier.WIKI_SEM_500
    """

    @classmethod
    def parse_question(cls, question):
        """
        Examples
        --------
        >>> Dataset.parse_question('january february march saturn')
        'january february march saturn'

        Parameters
        ----------
        question: str
            A splittable string with the group of words, outlier in last position

        Returns
        -------
        namedtuple

        """
        return ' '.join(question.strip().split())

    @classmethod
    def parse_file(cls, file_content):
        # a standard outlier detection dataset files contains first a cluster of words, one per line, then a blank line,
        # then a list of outliers for the cluster
        cluster = []
        outliers = []

        current = cluster
        for line in file_content:
            if not line.strip():
                current = outliers
                continue
            current.append(line.strip())

        return [' '.join(cluster) + ' ' + outlier for outlier in outliers]


_8_8_8 = Dataset("8-8-8", "../resources/en/outlier_detection/8-8-8")
WIKI_SEM_500 = Dataset("wiki-sem-500", "../resources/en/outlier_detection/wiki-sem-500-tokenized.zip")

ALL_DATASETS = [_8_8_8, WIKI_SEM_500]


# ####################################################################################################################
# EVALUATOR

class Evaluator(BaseEvaluator):
    """Evaluator to detect outliers in a group of words according to the given representation

    Parameters
    ----------
    representation: mangoes.Representation
        The Representation to use
    """

    def __init__(self, representation):
        self.representation = representation

    def predict(self, data):
        """Given a group of words or a set of group of words, predict the "outlier position" within each group

        The "outlier position" (OP) refers to [1]_ :

        Given a set W of n + 1 words, OP is defined as the position of the outlier w_{n+1} according to the compactness
        score, which ranges from 0 to n (position 0 indicates the lowest overall score among all words in W, and
        position n indicates the highest overall score).

        References
        ----------
        .. [1] José Camacho-Collados and Roberto Navigli. Find the word that does not belong: A Framework for an
               Intrinsic Evaluation of Word Vector Representations. In Proceedings of the ACL Workshop on Evaluating
               Vector Space Representations for NLP, Berlin, Germany, August 12, 2016.

        Examples
        --------
        >>> # create a representation
        >>> import numpy as np
        >>> import mangoes
        >>> vocabulary = mangoes.Vocabulary(['january', 'february', 'march', 'pluto', 'mars', 'saturn'])
        >>> matrix = np.array([[1.0, 0.2], [0.9, 0.1], [1.1, 0.1], [0.3, 0.9], [0.2, 1.0], [0.1, 0.9]])
        >>> representation = mangoes.Embeddings(vocabulary, matrix)
        >>> # predict
        >>> import mangoes.evaluation.outlier
        >>> evaluator = mangoes.evaluation.outlier.Evaluator(representation)
        >>> evaluator.predict('january february march saturn')
        4
        >>> evaluator.predict(['january february march saturn', 'pluto saturn march'])
        {'january february march saturn': 4, 'pluto saturn march': 3}

        Parameters
        ----------
        data: str or iterable of str

        Returns
        -------
        int or dict
            If a string was given, the outlier position according to the compactness score.
            If a list of string was given, a dict with strings as keys and outlier positions as values

        """
        if isinstance(data, str):
            compactness_scores = []
            group = data.split()
            for w in group:
                all_except_w = self.representation.matrix[[self.representation.words.index(word)
                                                           for word in group
                                                           if word is not w]]
                all_except_w = mangoes.utils.arrays.Matrix.factory(all_except_w)
                compactness_scores.append(_pseudo_inversed_compactness_score(all_except_w, self.representation[w]))
            sorted_indices = reversed(np.asarray(compactness_scores, dtype=float).argsort())
            return list(sorted_indices).index(len(group) - 1) + 1
        else:
            return {cluster: self.predict(cluster) for cluster in data}


def _pseudo_inversed_compactness_score(vectors, w):
    similarities = mangoes.utils.metrics.pairwise_cosine_similarity(w, vectors)
    return similarities.sum()


# ####################################################################################################################
# EVALUATION

class Evaluation(BaseEvaluation):
    """
    Examples
    --------
    >>> # create a representation
    >>> import numpy as np
    >>> import mangoes
    >>> vocabulary = mangoes.Vocabulary(['january', 'february', 'march', 'pluto', 'mars', 'saturn'])
    >>> matrix = np.array([[1.0, 0.2], [0.9, 0.1], [1.1, 0.1], [0.3, 0.9], [0.2, 1.0], [0.1, 0.9]])
    >>> representation = mangoes.Embeddings(vocabulary, matrix)
    >>> import mangoes.evaluation.outlier
    >>> # evaluate
    >>> dataset = Dataset("test", ['january february march pluto', 'mars saturn pluto march'])
    >>> evaluation = mangoes.evaluation.outlier.Evaluation(representation, dataset)
    >>> print(evaluation.get_score())
    Score(opp=1.0, accuracy=1.0, nb=2)
    >>> print(evaluation.get_report()) # doctest: +NORMALIZE_WHITESPACE
                                                                Nb questions         OPP    accuracy
    ================================================================================================
    test                                                                 2/2     100.00%     100.00%
    ------------------------------------------------------------------------------------------------
    """
    _Score = namedtuple("Score", "opp accuracy nb")

    class _Report(PrintableReport):
        HEADER = [(("Nb questions", "OPP", "accuracy"), (3, 3, 3))]
        PREDICTION_HEADER = [(("outlier position",), (6,))]

        def _print_score(self, score):
            if score.nb:
                return "{:>{width}}{:>{width}}".format("{:.2%}".format(score.opp), "{:.2%}".format(score.accuracy),
                                                       width=3 * self.COL)
            else:
                return "{:>{width}}{:>{width}}".format('NA', 'NA', width=3 * self.COL)

        def _print_prediction(self, question, indent):
            outlier_position = self.evaluation.predictions[question]

            line = "{:{width}}".format('    ' * indent + question, width=(self.LINE_LENGTH - 1 * self.COL))
            line += "{:>{width}}".format(outlier_position, width=1 * self.COL)
            return line + "\n"

    def __init__(self, representation, *datasets, lower=True):
        super(Evaluation, self).__init__(Evaluator(representation), *datasets, lower=lower)

    def _filter_list_of_questions(self, list_of_questions):
        subset_clusters = []
        unique_clusters = set()
        nb_oov = 0
        nb_duplicates = 0

        for group in list_of_questions:
            group = group.lower() if self.lower else group

            if all(w in self.evaluator.representation.words for w in group.split()):
                subset_clusters.append(group)

                if group in unique_clusters:
                    nb_duplicates += 1
                else:
                    unique_clusters.add(group)
            else:
                nb_oov += 1

        result = self._FilteredSubset(subset_clusters, len(list_of_questions), nb_oov, nb_duplicates)
        return result, unique_clusters, unique_clusters

    def _score(self, clusters):
        if not clusters:
            return np.nan, np.nan

        predictions = [self.predictions[cluster] for cluster in clusters]
        opp_score = sum([op / len(c.split()) for c, op in zip(clusters, predictions)]) / len(clusters)

        outlier_detections = [op == len(c.split()) for op, c in zip(predictions, clusters)]
        accuracy = sum(outlier_detections) / len(predictions)

        return opp_score, accuracy



# -*- coding: utf-8 -*-
"""Utility functions to save mangoes objects"""
import os
import tempfile
import zipfile
import json


def save(obj, attributes_to_persist, path, metadata=None):
    """Save the object at the given path

    Parameters
    ----------
    obj
        object to persists
    attributes_to_persist: list
        list o the names of the attributes of the object to persist
    path: str
        path to a folder or an archive where the object will be stored
    metadata: dict
        metadata to store in a .metadata file with the objects

    Returns
    -------
    str
        path to a folder or an archive
    """
    attributes_names = [name for (name, _) in attributes_to_persist]
    _, file_extension = os.path.splitext(path)

    if file_extension == ".zip":
        save_func = _save_as_archive
    else:
        save_func = _save_in_folder
    save_func(obj, path, attributes_names, metadata)


def load(cls, attributes_to_persist, path):
    """Load an object from the given path

    Parameters
    ----------
    path: str
        path to a folder or an archive

    Returns
    -------
    object of type `cls`
    """
    _, file_extension = os.path.splitext(path)

    if file_extension == ".zip":
        load_func = _load_from_archive
    else:
        load_func = _load_from_folder
    return load_func(cls, path, attributes_to_persist)


def _save_in_folder(obj, path, attributes, metadata=None):
    """Save the object's attributes as files in the local folder corresponding to the path.

    :param obj: object to store
    :param path: string, the path the folder where the embedding files will be stored.
    :param attributes: list of attributes to save.
    :return: None
    """
    if not os.path.exists(path):
        os.makedirs(path)

    if metadata:
        with open(os.path.join(path, ".metadata"), mode="w") as metadata_file:
            metadata_file.write(json.dumps(metadata, indent=4))

    for attribute in attributes:
        obj.__getattribute__(attribute).save(path, name=attribute)


def _load_from_folder(cls, path, attributes):
    """Load an object instance from the files in the specified folder.

    :param cls: class of the object to load
    :param path: string, path to the folder where the object instance files are located
    :param attributes: list of tuples defining attributes to load (name, cls)
    :return: an instance of 'cls'
    """
    args = []

    for attribute_name, attribute_class in attributes:
        args.append(attribute_class.load(path, attribute_name))

    if os.path.exists(os.path.join(path, ".metadata")):
        with open(os.path.join(path, ".metadata")) as metadata_file:
            args.append(json.loads(metadata_file.read()))

    return cls(*args)


def _save_as_archive(obj, path, attributes, metadata=None):
    """Save the object as an uncompressed zip archive.

    :param obj: object to store
    :param path: the path to the archive where the files will be stored.
    :param attributes: list of attributes to save.
    :return: None
    """
    with zipfile.ZipFile(path, 'w', zipfile.ZIP_STORED) as zip_file:
        with tempfile.TemporaryDirectory() as tmp_dir:
            for attribute in attributes:
                tmp_file = obj.__getattribute__(attribute).save(tmp_dir, name=attribute)
                zip_file.write(tmp_file, arcname=os.path.basename(tmp_file))

            if metadata:
                zip_file.writestr(".metadata", json.dumps(metadata, indent=4))


def _load_from_archive(cls, path, attributes):
    """Load an object from a zip archive file.

    :param cls: class of the object to load
    :param path: string, path to the archive where the object instance files are stored
    :param attributes: list of tuples defining attributes to load (name, cls)
    :return: an instance of 'cls'
    """

    with tempfile.TemporaryDirectory() as tmpdirname:
        with zipfile.ZipFile(path, 'r', zipfile.ZIP_STORED) as zip_file:
            zip_file.extractall(tmpdirname)
        return _load_from_folder(cls, tmpdirname, attributes)



# -*- coding: utf-8 -*-
"""Module to access to available datasets and create new ones.

Datasets available in this module :

* `WS353` for the `WordSim353 <http://alfonseca.org/eng/research/wordsim353.html>`_ dataset
  (Finkelstein et al., 2002) [1]_.\n
  Also partitioned by [2]_ into :
    * `WS_SIM` : WordSim Similarity
    * `WS_REL` : WordSim Relatedness
* `RG65` for Rubenstein and Goodenough (1965) dataset [3]_
* `RAREWORD` for the Luong et al.'s (2013)
  `Rare Word (RW) Similarity Dataset <https://nlp.stanford.edu/~lmthang/morphoNLM/>`_ [4]_
* `MEN` for the Bruni et al.'s (2012) MEN dataset [5]_
* `MTURK` for the `Radinsky et al.'s (2011) Mechanical Turk dataset` [6]_
* `SIMLEX` for the Hill et al.'s (2016) SimLex-999 dataset [7]_
* `GOOGLE` for the Mikolov et al.'s (2013) Google dataset [8]_ .\n
  Also partitionned into :
    * `GOOGLE_SEMANTIC` for semantic analogies
    * `GOOGLE_SYNTACTIC` for syntactic analogies
* `MSR` for the Mikolov et al.'s (2013) Microsoft Research dataset [9]_
* `OD_8_8_8` [10]_
* `WIKI_SEM_500` [11]_

Warnings
--------
The Simlex dataset is not compatible with this version of mangoes

References
----------
.. [1] Finkelstein, L., Gabrilovich, E., Matias, Y., Rivlin, E., Solan, Z., Wolfman, G., & Ruppin, E. (2001, April).
       Placing search in context: The concept revisited. In Proceedings of the 10th international conference on World
       Wide Web (pp. 406-414). ACM.
.. [2] Eneko Agirre, Enrique Alfonseca, Keith Hall, Jana Kravalova, Marius Pasca, Aitor Soroa, A Study on Similarity
       and Relatedness Using Distributional and WordNet-based Approaches, In Proceedings of NAACL-HLT 2009.
.. [3] Rubenstein, Herbert, and John B. Goodenough. Contextual correlates of synonymy. Communications of the ACM,
       8(10):627–633, 1965.
.. [4] Luong, T., Socher, R., & Manning, C. D. (2013, August). Better word representations with recursive neural
       networks for morphology. In CoNLL (pp. 104-113).
.. [5] Bruni, E., Boleda, G., Baroni, M., & Tran, N. K. (2012, July). Distributional semantics in technicolor. In
       Proceedings of the 50th Annual Meeting of the Association for Computational Linguistics: Long Papers-Volume 1
       (pp. 136-145). Association for Computational Linguistics.
.. [6] Radinsky, K., Agichtein, E., Gabrilovich, E., & Markovitch, S. (2011, March). A word at a time: computing word
       relatedness using temporal semantic analysis. In Proceedings of the 20th international conference on
       World wide web (pp. 337-346). ACM.
.. [7] Hill, F., Reichart, R., & Korhonen, A. (2016). Simlex-999: Evaluating semantic models with (genuine) similarity
       estimation. Computational Linguistics.
.. [8] Mikolov, T., Chen, K., Corrado, G., & Dean, J. (2013). Efficient estimation of word representations in vector
       space. arXiv preprint arXiv:1301.3781.
.. [9] Mikolov, T., Yih, W. T., & Zweig, G. (2013, June). Linguistic regularities in continuous space word
       representations. In hlt-Naacl (Vol. 13, pp. 746-751).
.. [10] José Camacho-Collados and Roberto Navigli. Find the word that does not belong: A Framework for an Intrinsic
        Evaluation of Word Vector Representations. In Proceedings of the ACL Workshop on Evaluating Vector Space
        Representations for NLP, Berlin, Germany, August 12, 2016.
.. [11]
"""

import warnings
warnings.warn("The dataset module is deprecated and will be removed in next release: use mangoes.evaluation package "
              "instead", DeprecationWarning)

import os.path
import zipfile
from collections import defaultdict, namedtuple

import pkg_resources

import mangoes.utils.exceptions
from mangoes.constants import ENCODING

_Question = namedtuple("Question", ["question", "gold"])


def _default_question_parser(question):
    return _Question(*question.rsplit(maxsplit=1))


def _outlier_detection_parser(question):
    return _Question(question, question.split()[-1])


class Dataset:
    """Base class to create an evaluation dataset.


    Parameters
    ----------
    content: str or list or dict
        Content of the dataset.Can be :
            - a path to a file containing the questions
            - a path to a directory containing such files : each subdirectory and each file will be
            considered as subsets.
            - a list of questions (each question is a string)
            Ex with analogies : ['king queen man woman', 'boy girl man woman', ...]
            - a dictionary containing list of questions structured in subsets.
            Ex : {"root": {"subset1":[...], "subset2":[...]}
    name: str
        name of the dataset
    lower: boolean
        True if the questions have to be lowercased

    Attributes
    ----------
    name
    lower
    subsets_to_questions: dict
        dictionary where keys are the name of the subsets and values are list of questions or nested subsets
    questions_to_subsets: dict
        dictionary where keys are the questions and values are the list of the subsets they belong to

    Examples
    --------
    >>> dataset = mangoes.dataset.Dataset(['a b c d', 'e f g h'], name="My dataset")
    >>> dataset.subsets_to_questions
    {'My dataset': ['a b c d', 'e f g h']}
    >>> dataset.questions_to_subsets
    {'a b c d': ['/My dataset'], 'e f g h': ['/My dataset']}

    >>> dataset = mangoes.dataset.Dataset({"subset1": ['a b c d','e f g h'], "subset2": ['a b c d']}, name="My dataset")
    >>> dataset.subsets_to_questions
    {'My dataset': {'subset1': ['a b c d', 'e f g h'], 'subset2': ['a b c d']}}
    >>> dataset.questions_to_subsets
    {'a b c d': {'/My dataset', '/My dataset/subset1', '/My dataset/subset2'},
     'e f g h': {'/My dataset', '/My dataset/subset1'}}

    """

    def __init__(self, content, name="", language="en", lower=True, question_parser=_default_question_parser):
        self.name = name
        self.language = language
        self.lower = lower
        self.parse_question = question_parser

        self.__questions_to_subsets = {}
        self.__subsets_to_questions = {}
        if isinstance(content, str):
            if not name:
                self.name = _get_name(content)
            content = self._read_from(content)
        else:
            content = {self.name: content}

        if self.lower:
            content = _lower_subset(content)

        self.subsets_to_questions = content

    @property
    def subsets_to_questions(self):
        return self.__subsets_to_questions

    @property
    def questions_to_subsets(self):
        return self.__questions_to_subsets

    @subsets_to_questions.setter
    def subsets_to_questions(self, subsets_to_questions):
        self.__subsets_to_questions = subsets_to_questions
        self.__questions_to_subsets = dict(self._flatten(self.__subsets_to_questions))

    def get_questions_and_gold(self, subset=None):
        """Returns a list of tuples with questions and expected answers

        Examples
        --------
        >>> dataset = Dataset({"My dataset": {"subset1": ['a b c d', 'e f g h'],
        >>>                                   "subset2": ['a b c d']}})
        >>> dataset.get_questions_and_gold()
        [Question(question='e f g', gold='h'), Question(question='a b c', gold='d')]
        >>> dataset.get_questions_and_gold("/My dataset/subset2")
        ['a b c d']

        Parameters
        ----------
        subset: str or None
            if None, return all the questions in the dataset, else, only the questions of the given subset.

        Returns
        -------
        list of tuples (question, gold)

        """
        if not subset:
            return [self.parse_question(q) for q in self.questions_to_subsets]
        return [self.parse_question(q) for q in self.questions_to_subsets
                if subset in self.questions_to_subsets[q]]

    @classmethod
    def _flatten(cls, subsets, prefix="", lower=True):
        result = defaultdict(set)

        for subset_name, subset_data in subsets.items():
            subset_path = prefix.split("/") + [subset_name]

            if isinstance(subset_data, dict):
                for question, list_of_subsets in cls._flatten(subset_data, prefix="/".join(subset_path)).items():
                    if lower:
                        question = question.lower()
                    result[question].update(list_of_subsets)

            else:
                for question_w_gold in subset_data:
                    if lower:
                        question_w_gold = question_w_gold.lower()
                    if subset_path:
                        result[question_w_gold].update(["/".join(subset_path[:i + 1])
                                                        for i in range(1, len(subset_path))
                                                        if subset_path[:i + 1]])
        return result

    @classmethod
    def _read_from(cls, path):
        if pkg_resources.resource_exists(__name__, path):
            path = pkg_resources.resource_filename(__name__, path)

        if os.path.isdir(path):
            return cls._read_from_dir(path)
        elif zipfile.is_zipfile(path):
            return cls._read_from_zip(path)
        else:
            try:
                return cls._read_from_file(path)
            except FileNotFoundError:
                raise mangoes.utils.exceptions.ResourceNotFound(path=path)

    @classmethod
    def _read_from_file(cls, path):
        with open(path, encoding=ENCODING) as file_content:
            return {_get_name(path): [line.strip() for line in file_content]}

    @classmethod
    def _read_from_dir(cls, path):
        questions = {}
        for subset in [os.path.join(path, subpath) for subpath in os.listdir(path)]:
            subset_name = _get_name(subset)
            if not subset_name.startswith('.'):
                content = cls._read_from(subset)
                questions[subset_name] = content[subset_name]
        return {_get_name(path): questions}

    @classmethod
    def _read_from_zip(cls, path):
        questions = {}
        with zipfile.ZipFile(path, 'r') as zip_file:
            for filename in zip_file.namelist():
                content = [line.strip() for line in zip_file.read(filename).decode().split('\n') if line.strip()]
                if content:
                    questions[_get_name(filename)] = content
        return {_get_name(path): questions}


class OutlierDetectionDataset(Dataset):
    """Base class for dataset for Outlier Detection task

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.parse_question = _outlier_detection_parser

    @classmethod
    def _read_from_file(cls, path):
        with open(path, encoding=ENCODING) as file_content:
            cluster, outliers = cls._parse_cluster_and_outliers(file_content)
            return {_get_name(path): [cluster + " " + outlier for outlier in outliers]}

    @classmethod
    def _read_from_zip(cls, path):
        questions = {}
        with zipfile.ZipFile(path, 'r') as zip_file:
            for filename in zip_file.namelist():
                cluster, outliers = cls._parse_cluster_and_outliers(zip_file.read(filename).decode().split('\n'))
                if outliers:
                    questions[_get_name(filename)] = [cluster + " " + outlier for outlier in outliers]
        return {_get_name(path): questions}

    @classmethod
    def _parse_cluster_and_outliers(cls, cluster_and_outliers):
        cluster = []
        outliers = []

        current = cluster
        for line in cluster_and_outliers:
            if not line.strip():
                current = outliers
                continue
            current.append(line.strip())
        return " ".join(cluster), outliers


def _get_name(content):
    return os.path.splitext(os.path.basename(content))[0]


def nb_questions(subset):
    """Returns the total number of questions in a subset of a Dataset

    Parameters
    ----------
    subset: dict
        subset of a Dataset

    Returns
    -------
    int

    """
    if not isinstance(subset, dict):
        return len(subset)
    return sum([nb_questions(v) for v in subset.values()])


def _lower_subset(subset):
    if isinstance(subset, list):
        return [y.lower() for y in subset]
    for key in subset.keys():
        subset[key] = _lower_subset(subset[key])
    return subset


WS353 = "ws353"
WS353_RELATEDNESS = "ws353_relatedness"
WS353_SIMILARITY = "ws353_similarity"
MEN = "men"
MTURK = "mturk"
RAREWORD = "rareword"
RG65 = "rg65"
SIMLEX = "simlex"
GOOGLE = "google"
GOOGLE_SEMANTIC = "google_semantic"
GOOGLE_SYNTACTIC = "google_syntactic"
MSR = "msr"
ANALOGY = "analogy"
SIMILARITY = "similarity"
OUTLIER_DETECTION = "outlier_detection"
OD_8_8_8 = "8-8-8"
WIKI_SEM_500 = "wiki_sem_500"

AVAILABLE_DATASETS = {"en": {WS353: ("resources/en/similarity/wordsim353.txt", Dataset),
                             WS353_RELATEDNESS: ("resources/en/similarity/ws353_relatedness.txt", Dataset),
                             WS353_SIMILARITY: ("resources/en/similarity/ws353_similarity.txt", Dataset),
                             MEN: ("resources/en/similarity/men.txt", Dataset),
                             MTURK: ("resources/en/similarity/mturk.txt", Dataset),
                             RAREWORD: ("resources/en/similarity/rareword.txt", Dataset),
                             RG65: ("resources/en/similarity/rg65.txt", Dataset),
                             # SIMLEX: ("resources/en/similarity/SimLex-999.txt", Dataset),
                             GOOGLE: ("resources/en/analogy/google", Dataset),
                             GOOGLE_SEMANTIC: ("resources/en/analogy/google/semantic", Dataset),
                             GOOGLE_SYNTACTIC: ("resources/en/analogy/google/syntactic", Dataset),
                             MSR: ("resources/en/analogy/msr/syntactic", Dataset),
                             ANALOGY: ("resources/en/analogy", Dataset),
                             SIMILARITY: ("resources/en/similarity", Dataset),
                             OUTLIER_DETECTION: ("resources/en/outlier_detection", OutlierDetectionDataset),
                             OD_8_8_8: ("resources/en/outlier_detection/8-8-8", OutlierDetectionDataset),
                             WIKI_SEM_500: (
                                 "resources/en/outlier_detection/wiki-sem-500-tokenized.zip", OutlierDetectionDataset)},
                      "fr": {RG65: ("resources/fr/similarity/rg65-fr.txt", Dataset),
                             SIMILARITY: ("resources/fr/similarity", Dataset), }
                      }


def load(dataset_name, language="en", lower=True):
    """Loads a dataset from the AVAILABLE_DATASETS

    Parameters
    ----------
    dataset_name: str
        the name of the dataset, must be in AVAILABLE_DATASETS
    language: {'en', 'fr'}
        Code of the language of the dataset (default = 'en')
    lower: boolean
        whether the questions of the dataset should be lowercased

    Returns
    -------
    mangoes.Dataset
    """
    if dataset_name == SIMLEX:
        raise NotImplementedError("The Simlex dataset is not yet compatible")
        # return _load_simlex(lower=lower)
    if not language:
        language = "en"
    path, handler = AVAILABLE_DATASETS[language][dataset_name]
    return handler(path, language=language, lower=lower)

# def _load_simlex(lower=True):
#     TODO : adapt to use the simlex dataset
# def parse_question(question):
#     word1, word2, _, gold, *_ = question.split()
#     return Dataset.Question(question=' '.join([word1, word2]), gold=gold)
#
# path, handler = AVAILABLE_DATASETS[SIMLEX]
# return handler(path, lower=lower, question_parser=parse_question)



# -*-coding:utf-8 -*
"""This module provides different ways to derive a phrase vector from the vectors of its parts.

Given two vectors :math:`u` and :math:`v` representing two words *w1* and *w2*, composition methods can be applied to
derive a new vector :math:`p` representing the phrase *'w1 w2'* by combining :math:`u` and :math:`v`.

This module provides classes to learn parameters for various compositional models and apply them to predict new vectors.

These models are :

- additive model [#3]_: :math:`p` is obtained by a (weighted) sum of :math:`u` and :math:`v` :

    .. math::
        \mathbf{p = \\alpha u + \\beta v}

  These weights can be learned with :class:`.AdditiveComposer`

- multiplicative [#3]_: :math:`p` is obtained by component-wise multiplication of :math:`u` and :math:`v` :

    .. math::
        \mathbf{p = u \odot v}

  This method has no parameter to learn but :class:`.MultiplicativeComposer` is provided here to be
  compared with other ones.

- dilation model [#2]_ [#3]_: :math:`p` is obtained by calculating the dot products of :math:`u.u` and :math:`u.v`
  and stretching :math:`v` by a factor :math:`\lambda` in the direction of :math:`u` :
    .. math::
        \mathbf{p = (u.u)v + (\lambda - 1)(u.v)}

  :math:`\lambda` can be learned with :class:`.DilationComposer`

- full additive [#4]_: an extension of the additive model where the two n-dimension input vectors are multiplied by
  two n x n weight matrices :
    .. math::
        \mathbf{p = Au + Bv}
  :math:`A` and :math:`B` can be learned with :class:`.FullAdditiveComposer`

- lexical function [#5]_: *w1* is seen as a function and represented as a matrix :math:`U` and :math:`p` is the product
  of this matrix and :math:`v`
    .. math::
        \mathbf{p = Uv}
  :math:`U` can be learned with :class:`.LexicalComposer`

References
----------
.. [#1] Boleda, G., Baroni, M., & McNally, L. (2013). Intensionality was only alleged: On adjective-noun composition in
        distributional semantics. In Proceedings of the 10th International Conference on Computational Semantics
        (IWCS 2013)–Long Papers (pp. 35-46).
.. [#2] Clark, S., Coecke, B., & Sadrzadeh, M. (2008). A compositional distributional model of meaning. In Proceedings
        of the Second Quantum Interaction Symposium (QI-2008) (pp. 133-140).
.. [#3] Mitchell, J., & Lapata, M. (2010). Composition in distributional models of semantics. Cognitive science,
        34(8), 1388-1429.
.. [#4] Guevara, E. (2010, July). A regression model of adjective-noun compositionality in distributional semantics.
        In Proceedings of the 2010 Workshop on GEometrical Models of Natural Language Semantics (pp. 33-37). Association
        for Computational Linguistics.
.. [#5] Baroni, M., & Zamparelli, R. (2010, October). Nouns are vectors, adjectives are matrices: Representing
        adjective-noun constructions in semantic space. In Proceedings of the 2010 Conference on Empirical Methods in
        Natural Language Processing (pp. 1183-1193). Association for Computational Linguistics.

"""

import abc
import logging

import scipy
import numpy as np
import sklearn.cross_decomposition

import mangoes
import mangoes.utils.exceptions

_logger = logging.getLogger(__name__)


class _Composer:
    """Base class for composers"""
    def __init__(self, representation, init=None):
        if representation.matrix.is_sparse:
            # TODO: implement for sparse matrices
            return NotImplementedError
        self.representation = representation
        self.init = init
        self.params = None

    @abc.abstractmethod
    def compose(self, *args):
        pass

    def predict(self, u, v):
        u = self.representation[u]
        v = self.representation[v]
        return self.compose(u, v, *self.params)


class _ScalarComposer(_Composer):
    """Base class for composers learning scalars, using least square regression"""
    def residual(self, parameters, observed, adjectives, nouns):
        return [scipy.spatial.distance.euclidean(o, self.compose(u, v, *parameters))
                for o, u, v in zip(observed, adjectives, nouns)]

    def _fit(self, bigrams=None):
        if not bigrams:
            bigrams = self.representation.words.get_bigrams()
        bigrams = mangoes.Vocabulary(bigrams)

        x = [self.representation[b[0]] for b in bigrams]
        y = [self.representation[b[1]] for b in bigrams]

        observed_vectors = [self.representation[b] for b in bigrams]

        result = scipy.optimize.least_squares(self.residual, self.init, args=(observed_vectors, x, y))
        self.params = result.x

    def transform(self, bigram):
        return self.compose(*bigram)


class AdditiveComposer(_ScalarComposer):
    """Compose phrase vectors using the additive model

    Given two vectors :math:`u` and :math:`v` representing two words *w1* and *w2*, the Additive Model derives
    a new vector :math:`p` representing the phrase *'w1 w2'* by combining :math:`u` and :math:`v` by a (weighted) sum
    of :math:`u` and :math:`v` :

    .. math::
        \mathbf{p = \\alpha u + \\beta v}

    This class learn these weights from a Representation then apply them to predict new vectors

    Examples
    --------
    >>> import numpy
    >>> import mangoes.composition
    >>> colors = ['white', 'black', 'green', 'red']
    >>> nouns = ['dress', 'rabbit', 'flag']
    >>> adj_nouns = ['white dress', 'black dress', 'red flag', 'green flag', 'white rabbit']
    >>> vocabulary = mangoes.Vocabulary(colors + nouns + adj_nouns)
    >>> matrix = numpy.random.random((12, 5))
    >>> embeddings = mangoes.Embeddings(vocabulary, matrix)
    >>> additive_composer = mangoes.composition.AdditiveComposer(embeddings)
    >>> additive_composer.fit()
    >>> green_rabbit = additive_composer.predict('green', 'rabbit')

    Parameters
    ----------
    representation: mangoes.Representation
        Representation from which the weights will be learned. So, the vocabulary represented should contains bigrams
        where both part are also parts of the vocabulary.

    init: (float, float)
        initial values for alpha and beta. Default = (1, 1)

    Attributes
    ----------
    alpha:
        Learned weight applied to first words
    beta:
        Learned weight applied to second words
    """

    def __init__(self, representation, init=(1,1)):
        super().__init__(representation, init)

    @property
    def alpha(self):
        return self.params[0]

    @property
    def beta(self):
        return self.params[1]

    @staticmethod
    def compose(u, v, a, b):
        return a * u + b * v

    def fit(self, bigrams=None):
        """Fit model to data

        Parameters
        ----------
        bigrams: list
            If bigrams = None (default), parameters will be learned from all the bigrams found in the vocabulary of the
            Representation. But a list of bigrams can be provided here : it has to be a subset of the vocabulary of the
            Representation.
        """
        self._fit(bigrams)


class DilationComposer(_ScalarComposer):
    """Compose phrase vectors using the dilation model

    Given two vectors :math:`u` and :math:`v` representing two words *w1* and *w2*, the dilation model derives
    a new vector :math:`p` representing the phrase *'w1 w2'* by calculating the dot products of :math:`u.u` and
    :math:`u.v` and stretching :math:`v` by a factor :math:`\lambda` in the direction of :math:`u` :

    .. math::
        \mathbf{p = (u.u)v + (\lambda - 1)(u.v)}

    This class learn :math:`\lambda` from a Representation then apply the dilation to predict new vectors

    Examples
    --------
    >>> import numpy
    >>> import mangoes.composition
    >>> colors = ['white', 'black', 'green', 'red']
    >>> nouns = ['dress', 'rabbit', 'flag']
    >>> adj_nouns = ['white dress', 'black dress', 'red flag', 'green flag', 'white rabbit']
    >>> vocabulary = mangoes.Vocabulary(colors + nouns + adj_nouns)
    >>> matrix = numpy.random.random((12, 5))
    >>> embeddings = mangoes.Embeddings(vocabulary, matrix)
    >>> dilation_composer = mangoes.composition.AdditiveComposer(embeddings)
    >>> dilation_composer.fit()
    >>> green_rabbit = dilation_composer.predict('green', 'rabbit')

    Parameters
    ----------
    representation: mangoes.Representation
        Representation from which the dilation factor will be learned. So, the vocabulary represented should contains
        bigrams where both part are also parts of the vocabulary.

    init: float
        initial value for lambda_. Default = 0.5

    Attributes
    ----------
    lambda_ : float
        The dilation factor
    """
    def __init__(self, representation, init=0.5):
        super().__init__(representation, init)

    @property
    def lambda_(self):
        return self.params[0]

    @staticmethod
    def compose(u, v, lambda_):
        return (lambda_ - 1) * u.dot(v.T) * u + u.dot(u.T) * v

    def fit(self, bigrams=None):
        """Fit model to data

        Parameters
        ----------
        bigrams: list
            If bigrams = None (default), lambda_ will be learned from all the bigrams found in the vocabulary of the
            Representation. But a list of bigrams can be provided here : it has to be a subset of the vocabulary of the
            Representation.
        """
        self._fit(bigrams)


class MultiplicativeComposer(_Composer):
    """Compose phrase vectors using the multiplicative model

    Given two vectors :math:`u` and :math:`v` representing two words *w1* and *w2*, the multiplicative model derives
    a new vector :math:`p` representing the phrase *'w1 w2'* by component-wise multiplication of :math:`u`
    and :math:`v` :

    .. math::
        \mathbf{p = u \odot v}

    This class has no parameter to learn but is provided with a fit() function to be easily compared with other models.

    Examples
    --------
    >>> import numpy
    >>> import mangoes.composition
    >>> colors = ['white', 'black', 'green', 'red']
    >>> nouns = ['dress', 'rabbit', 'flag']
    >>> adj_nouns = ['white dress', 'black dress', 'red flag', 'green flag', 'white rabbit']
    >>> vocabulary = mangoes.Vocabulary(colors + nouns + adj_nouns)
    >>> matrix = numpy.random.random((12, 5))
    >>> embeddings = mangoes.Embeddings(vocabulary, matrix)
    >>> multiplicative_composer = mangoes.composition.MultiplicativeComposer(embeddings)
    >>> multiplicative_composer.fit() # does nothing
    >>> green_rabbit = multiplicative_composer.predict('green', 'rabbit')

    Parameters
    ----------
    representation: mangoes.Representation
        Representation used to predict phrase vectors.

    """
    def __init__(self, representation):
        super().__init__(representation)

    @staticmethod
    def compose(u, v):
        return u * v

    def fit(self, bigrams=None):
        pass

    def predict(self, u, v):
        u = self.representation[u]
        v = self.representation[v]
        return self.compose(u, v)


class FullAdditiveComposer(_Composer):
    """Compose phrase vectors using the full additive model

    Given two vectors :math:`u` and :math:`v` representing two words *w1* and *w2*, the Full Additive Model derives
    a new vector :math:`p` representing the phrase *'w1 w2'* by multiplying :math:`u` and :math:`v`
    by two n x n weight matrices :

    .. math::
        \mathbf{p = Au + Bv}
    can be learned with :class:`.FullAdditiveComposer`

    This class learn :math:`A` and :math:`B` from a Representation, using PLSR, then apply them to predict new vectors

    Examples
    --------
    >>> import numpy
    >>> import mangoes.composition
    >>> colors = ['white', 'black', 'green', 'red']
    >>> nouns = ['dress', 'rabbit', 'flag']
    >>> adj_nouns = ['white dress', 'black dress', 'red flag', 'green flag', 'white rabbit']
    >>> vocabulary = mangoes.Vocabulary(colors + nouns + adj_nouns)
    >>> matrix = numpy.random.random((12, 5))
    >>> embeddings = mangoes.Embeddings(vocabulary, matrix)
    >>> composer = mangoes.composition.FullAdditiveComposer(embeddings)
    >>> composer.fit()
    >>> green_rabbit = composer.predict('green', 'rabbit')

    Parameters
    ----------
    representation: mangoes.Representation
        Representation from which the weights will be learned. So, the vocabulary represented should contains bigrams
        where both part are also parts of the vocabulary.

    Attributes
    ----------
    A: array
    B: array
        Learned weight applied to first and second words. The shape of the matrices is nxn where n is the
        dimension of the representation
    """
    def __init__(self, representation):
        super().__init__(representation)

    @staticmethod
    def compose(u, v, A, B):
        return A.dot(u) + B.dot(v)

    def fit(self, bigrams=None):
        """Fit model to data

        Parameters
        ----------
        bigrams: list
            If bigrams = None (default), parameters will be learned from all the bigrams found in the vocabulary of the
            Representation. But a list of bigrams can be provided here : it has to be a subset of the vocabulary of the
            Representation.
        """
        if not bigrams:
            bigrams = self.representation.words.get_bigrams()
        bigrams = mangoes.Vocabulary(bigrams)
        pls = sklearn.cross_decomposition.PLSRegression(n_components=self.representation.shape[1]*2, scale=False)

        U = [self.representation[b[0]] for b in bigrams]
        V = [self.representation[b[1]] for b in bigrams]
        observed_vectors = [self.representation[b] for b in bigrams]

        X = np.hstack((U, V))

        pls.fit(X, observed_vectors)
        self.params = pls

        self.A = pls.coef_[:self.representation.shape[1]].T
        self.B = pls.coef_[self.representation.shape[1]:].T


    def predict(self, u, v):
        u = self.representation[u]
        v = self.representation[v]
        return self.params.predict(np.hstack((u, v)).reshape(1, -1))[0]


class LexicalComposer(_Composer):
    """Compose phrase vectors using the lexical model

    Given a word *w1*, the Full Additive Model sees it as a function, represented as a matrix :math:`U` and applies
    it to another word *w2* represented as a vector :math:`v` to derives a new vector :math:`p` representing the
    phrase *'w1 w2'*

    .. math::
        \mathbf{p = Uv}

    This class learn :math:`U` for word *w1* from a Representation, using PLSR, then apply it to predict new vectors

    Examples
    --------
    >>> import numpy
    >>> import mangoes.composition
    >>> colors = ['green']
    >>> nouns = ['dress', 'hat', 'bean', 'lantern', 'rabbit']
    >>> adj_nouns = ['green dress', 'green bean', 'green hat', 'green lantern']
    >>> vocabulary = mangoes.Vocabulary(colors + nouns + adj_nouns)
    >>> matrix = numpy.random.random((10, 5))
    >>> embeddings = mangoes.Embeddings(vocabulary, matrix)
    >>> green = mangoes.composition.LexicalComposer(embeddings, 'green')
    >>> green.fit()
    >>> green_rabbit = green.predict('rabbit')

    Parameters
    ----------
    representation: mangoes.Representation
        Representation from which the weights will be learned. So, the vocabulary represented should contains bigrams
        where both part are also parts of the vocabulary.

    word: str
        The word to represent as a matrix

    n_components: int
        Number of components to keep in PLSR. If None, use 1/4 of the numbers of bigrams that will be used in fit.

    Attributes
    ----------
    U: array
        Learned matrix to represent `word`. The shape of the matrices is nxn where n is the
        dimension of the representation
    """
    def __init__(self, representation, word, n_components=None):
        super().__init__(representation)
        self.word = word
        self.n_components = n_components

    @staticmethod
    def compose(U, v):
        return U.dot(v)

    def fit(self, bigrams=None):
        """Fit model to data

        Parameters
        ----------
        bigrams: list
            If bigrams = None (default), parameters will be learned from all the bigrams found in the vocabulary of the
            Representation. But a list of bigrams can be provided here : it has to be a subset of the vocabulary of the
            Representation.
        """
        if not bigrams:
            bigrams = self.representation.words.get_bigrams()
        bigrams_with_word = [b for b in bigrams
                             if b[0] == self.word and b[1] in self.representation.words]

        if not self.n_components:
            self.n_components = len(bigrams_with_word) // 4
            if self.n_components == 0:
                self.n_components = min(len(bigrams_with_word), self.representation.shape[1])

        y = [self.representation[b[1]] for b in bigrams_with_word]
        observed_vectors = [self.representation[b] for b in bigrams_with_word]

        pls = sklearn.cross_decomposition.PLSRegression(n_components=self.n_components, scale=False)
        pls.fit(y, observed_vectors)

        self.params = pls

    def predict(self, v):
        v = self.representation[v]
        return self.params.predict(v.reshape(1, -1))[0]

    @property
    def U(self):
        return self.params.coef_.T



# -*- coding: utf-8 -*-
"""
Mangoes is a toolbox to produce and evaluate word embeddings.

Mangoes can be used to :

* Construct a words co-occurrence matrix from a Corpus
* Create Word Embeddings from this matrix
* Analyze a corpus (counting words, sentences, ...) and create vocabularies from it
* Evaluate word embeddings
* Generate context dependent embeddings from a pretrained language model.
* Train a BERT language model
* Fine-tune a BERT language model

"""

from mangoes.corpus import Corpus
from mangoes.dataset import Dataset
from mangoes.vocabulary import Vocabulary
from mangoes.base import Embeddings, CountBasedRepresentation, create_representation
import mangoes.modeling
import mangoes.context
import mangoes.counting
import mangoes.dataset
import mangoes.evaluate
import mangoes.reduction
import mangoes.weighting

__version__ = '2.0.0'



# -*- coding: utf-8 -*-

import logging
import multiprocessing
import multiprocessing.queues
import traceback

logger = logging.getLogger(__name__)


class _Traceback(object):
    def __init__(self, trace):
        self.trace = trace


class _Sentinel(object):
    pass


class DataParallel:
    """Utility class to parallelize a function to apply on iterable data

    Parameters
    ----------
    task: callable
        function that takes an iterable as first argument and returns a value that can be reduced with the 'reducer'
        parameter.
        Its signature should be : `task_function(jobs_iterator, *args, **kwargs)`

    reducer: callable
        function that takes two objects of the same type (same type as the output of the 'task' function), merge them
        and returns an object of the same type.
        Its signature should be : `reduce_function(current_output, new_output)`
    nb_workers: int
        number of subprocess to use (default=1)
    batch: int, default None
        if provided, yields the result of task and call the reducer after `batch` iterations

    """

    def __init__(self, task, reducer, nb_workers=1, batch=None):
        self.task = task
        self.reducer = reducer

        self.nb_workers = max(1, min(nb_workers, multiprocessing.cpu_count()))
        self.workers = []
        self.job_dispatch_worker = None

        self.batch = batch

        self.input_queue = multiprocessing.Queue(maxsize=2*nb_workers)
        self.output_queue = multiprocessing.Queue()

        self._logger = logging.getLogger("{}.{}".format(__name__, self.__class__.__name__))

    def run(self, data, *args, **kwargs):
        """Run the `task` function on `data` with `nb_workers` processes.

        Parameters
        ----------
        data: iterator
            iterator over data to process
        args
            arguments for `task` function
        kwargs
            keywords arguments for `task` function

        Returns
        -------
        the result of the merging of all the values returned by `task`
        """
        running_workers = self._start_task_workers(max(1, self.nb_workers - 1), *args, **kwargs)
        self._start_dispatching_worker(data)

        try:
            result = self._process_output_queue(running_workers)
        except Exception:
            self._terminate()
            raise

        self.job_dispatching_worker.join()
        logger.debug("'Job dispatcher' worker terminated.")
        return result

    def _process_output_queue(self, running_workers):
        final_result = None
        while running_workers > 0:
            # Listen for regular outputs or exception's traceback
            try:
                output = self.output_queue.get()

                if isinstance(output, _Traceback):
                    raise Exception(output.trace)

                worker_id, part_result = output

                if isinstance(part_result, _Sentinel):
                    self.workers[worker_id].join()
                    running_workers -= 1
                    logger.debug("'Mapper' worker n°{} terminated, {} running.".format(worker_id, running_workers))
                else:
                    if final_result is None:
                        final_result = part_result
                    else:
                        final_result = self.reducer(final_result, part_result)
            except multiprocessing.queues.Empty:
                pass
        return final_result

    def _start_task_workers(self, nb, *args, **kwargs):
        self.workers = [self._create_a_task_worker(worker_id, *args, **kwargs)
                        for worker_id in range(nb)]
        for worker in self.workers:
            worker.start()
        return nb

    def _start_dispatching_worker(self, jobs):
        self.job_dispatching_worker = multiprocessing.Process(target=self._create_queue,
                                                              args=(jobs,))
        self.job_dispatching_worker.daemon = True
        self.job_dispatching_worker.start()

    def _create_queue(self, jobs):
        try:
            for job in jobs:
                self.input_queue.put(job)

            for _ in range(self.nb_workers):
                self.input_queue.put(_Sentinel)
        except Exception:
            tb = traceback.format_exc()
            self.output_queue.put(_Traceback(tb))

    def _worker(self, worker_id, *args, **kwargs):
        try:
            for primary_output in self.task(iter(self.input_queue.get, _Sentinel), *args, **kwargs):
                output = (worker_id, primary_output)
                self.output_queue.put(output)
            self.output_queue.put((worker_id, _Sentinel()))
        except Exception:
            tb = traceback.format_exc()
            self.output_queue.put(_Traceback(tb))

    def _create_a_task_worker(self, worker_id, *args, **kwargs):
        if self.batch:
            kwargs['batch'] = self.batch
        worker = multiprocessing.Process(target=self._worker,
                                         args=[worker_id, *args], kwargs=kwargs)
        worker.daemon = True
        return worker

    def _terminate(self):
        try:
            for worker in self.workers:
                worker.terminate()
            self.job_dispatch_worker.terminate()
        except Exception:
            pass



# -*- coding: utf-8 -*-
""" Class to manage the words to be represented in embeddings or used as contexts.

"""

import logging
import os.path
import json
import collections

import mangoes.utils.decorators
import mangoes.utils.exceptions
from mangoes.constants import ENCODING

_logger = logging.getLogger(__name__)


Bigram = collections.namedtuple('Bigram', 'first second')


class Vocabulary:
    """List of words.

    Vocabulary encapsulates a mapping between words and their ids.
    A Vocabulary can be create from a collection of words.

    Parameters
    ----------
    source: list or dict
        List of words or dict where keys are words and values are their indices

    language: str (optional)

    entity: str or tuple (optional)
        if the words are annotated, attribute(s) of each word

    See Also
    ---------
    :func:`mangoes.corpus.Corpus.create_vocabulary`
    """

    FILE_HEADER_PREFIX = "_$"

    def __init__(self, source, language=None, entity=None):
        self._params = {"language": language, "entity": entity}

        self._index_word = []
        self.word_index = {}
        self._factory(source)

    def __len__(self):
        return len(self._index_word)

    def __eq__(self, other):
        return self._index_word == other._index_word

    def __ne__(self, other):
        return not self.__eq__(other)

    def __iter__(self):
        return self._index_word.__iter__()

    def __contains__(self, word):
        return word in self.word_index

    def __getitem__(self, index):
        return self._index_word[index]

    def __repr__(self):
        return 'Vocabulary(' + str(self.words) + ')'

    def copy(self):
        v = mangoes.Vocabulary(self.words.copy())
        v._params = self._params
        return v

    @property
    def language(self):
        return self._params["language"]

    @property
    def entity(self):
        return self._params["entity"]

    @property
    def params(self):
        return self._params

    @property
    def words(self):
        """Returns the words of the vocabulary as a list"""
        return self._index_word

    def get_bigrams(self):
        return [w for w in self.words if isinstance(w, Bigram)]

    def index(self, word):
        """Returns the index associated to the word"""
        try:
            return self.word_index[word]
        except KeyError:
            raise mangoes.utils.exceptions.OutOfVocabulary(value=word)

    def indices(self, sentence):
        """Convert words of the sentence to indices

        If a word isn't in the vocabulary, its index is replaced with -1

        Parameters
        ----------
        sentence: list of str

        Returns
        -------
        list of int
        """
        return [self.index(word) if word in self else -1 for word in sentence]
        # TODO: rename encode and deal with entities fields (with a token filter)
        #     filter_word_sentence = mangoes.vocabulary.create_tokens_filter(words_vocabulary.entity)
        #     filter_bigrams_sentence = mangoes.vocabulary.create_bigrams_filter(words_vocabulary.get_bigrams())

    def append(self, word):
        """Append the word to the vocabulary"""
        if word not in self.word_index:
            self.word_index[word] = len(self.words)
            self.words.append(word)
        return self.word_index[word]

    def extend(self, other, inplace=True, return_map=False):
        """Extend the vocabulary with words in other

        Parameters
        ----------
        other: list or Vocabulary
        inplace: boolean
            If False, create a new Vocabulary
        return_map: boolean
            If True, the mapping between the indices of the words from original to merged is returned

        Returns
        -------
        Vocabulary or (Vocabulary, dict)
            Returns the merged vocabulary and, if return_map is True, the mapping between the indices of the words
            from original to merged

        """
        indices_map = {}
        if inplace:
            merged = self
        else:
            merged = self.copy()

        for word in other:
            new_index = merged.append(word)
            indices_map[other.index(word)] = new_index

        if return_map:
            return merged, indices_map

        return merged

    def save(self, path, name="vocabulary"):
        """Save the vocabulary in a file.

        Parameters
        ----------
        path: str
            Local path to the directory where vocabulary should be written

        name: str
            Name of the file to create (without extension)

        Warnings
        ---------
        If the file already exists, it will be overwritten.
        """

        file_path = os.path.join(path, name + '.txt')
        with open(file_path, "w", encoding=ENCODING) as f:
            f.write(self.FILE_HEADER_PREFIX + json.dumps({k: v for k, v in self._params.items() if v},
                                                         allow_nan=False) + "\n")
            if self.entity and not isinstance(self.entity, str):
                f.write("\n".join([self._token_to_string(t[0]) + ' ' + self._token_to_string(t[1])
                                   if isinstance(t, Bigram)
                                   else self._token_to_string(t)
                                   for t in self._index_word]))
            else:
                f.write("\n".join([str(w[0]) + ' ' + str(w[1]) if isinstance(w, Bigram) else str(w) for w in self._index_word]))
        return file_path

    def _token_to_string(self, token):
        return '/'.join([getattr(token, field) for field in self.entity])

    @classmethod
    def load(cls, path, name):
        """Load the vocabulary from its associated file.

        Parameters
        -----------
        path: str
            Local path to the directory where vocabulary file is located

        name: str
            Name of the file (without extension)

        Returns
        --------
        Vocabulary
        """
        temp_words = []
        params = {}

        def parse_words(line):
            line = line.strip()
            if ' ' in line:
                return Bigram(*line.split(' '))
            else:
                return line

        def create_token_parser(token_fields):
            Token = collections.namedtuple('Token', token_fields)

            def parse_tokens(line):
                line = line.strip()
                if ' ' in line:
                    bigram = line.split(' ')
                    return Bigram(Token(*bigram[0].split('/')), Token(*bigram[1].split('/')))
                else:
                    return Token(*line.split('/'))  # simple token
            return parse_tokens

        with open(os.path.join(path, name + '.txt'), "r", encoding=ENCODING) as f:
            first_line = f.readline()
            if first_line.startswith(cls.FILE_HEADER_PREFIX):
                params = json.loads(first_line[len(cls.FILE_HEADER_PREFIX):].strip())
                if 'entity' in params and not isinstance(params['entity'], str):
                    parse_line = create_token_parser(params['entity'])
                else:
                    parse_line = parse_words
            else:
                parse_line = parse_words
                f.seek(0)

            for line in f:
                temp_words.append(parse_line(line))

        return Vocabulary(temp_words, **params)

    def _factory(self, words):
        if isinstance(words, self.__class__):
            self._params = words._params
            self._from_list(words._index_word)
        elif isinstance(words, dict):
            self._from_dict(words)
        elif isinstance(words, list):
            self._from_list(words)
        else:
            try:
                self._from_list(words.words)
            except:
                error_message = "{} can't be used as input to create a Vocabulary. " \
                                "A Vocabulary, a dict or a list is expected".format(type(words))
                raise mangoes.utils.exceptions.UnsupportedType(error_message)

    def _from_dict(self, word_index):
        words_list = sorted(word_index, key=word_index.get)
        self._from_list(words_list)

    def _from_list(self, words):
        self._check_bigrams(words)
        self._index_word = words
        self.word_index = {word: index for index, word in enumerate(self._index_word)}

    def _check_bigrams(self, words):
        """Detect and replace bigrams with a Vocabulary.Bigram instance"""
        for i, w in enumerate(words):
            if isinstance(w, str) and ' ' in w:
                # a two words string
                words[i] = Bigram(*w.split())
            elif isinstance(w, tuple) and len(w) == 2:
                # a tuple of length 2 can be :
                # - a tuple of string -> a bigram
                # - a tuple of Tokens -> a bigram
                # - a Token with 2 fields : not a bigram
                if not getattr(w, '_fields', None):
                    # not a namedtuple so (probably ?) not a Token
                    words[i] = Bigram(*w)

    def __getstate__(self):
        state = self.__dict__.copy()
        if self.entity and not isinstance(self.entity, str):
            state['_index_word'] = [tuple(t) for t in self.words]
            del state['word_index']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        if self.entity and not isinstance(self.entity, str):
            Token = collections.namedtuple('Token', self.entity)
            self._from_list([Token(*t) for t in self._index_word])


class DynamicVocabulary(Vocabulary):
    """Extensible list of words.

    A DynamicVocabulary can be created from a collection of words or empty and each new encountered word will be
    added to it either explicitly (with :func:`.add()` or implicitly when testing if the word is in the vocabulary
    (always returns True) or getting its index

    Examples
    --------
    >>> v = mangoes.vocabulary.DynamicVocabulary()
    >>> print(v.words)
    []
    >>> v.append('a')
    0
    >>> print(v.words)
    ['a']
    >>> 'b' in v
    True
    >>> v.index('b')
    1
    >>> v.index('c')
    2


    Parameters
    ----------
    source: list or dict (optional)
        List of words or dict where keys are words and values are their indices

    language: str (optional)

    entity: str or tuple (optional)
        if the words are annotated, attribute(s) of each word

    See Also
    ---------
    :func:`mangoes.corpus.Corpus.create_vocabulary`
    """
    def __init__(self, source=None, *args, **kwargs):
        if not source:
            source = []
        super().__init__(source, *args, **kwargs)

    def index(self, word):
        """Returns the index associated to the word, adding it to the vocabulary if not yet"""
        try:
            return self.word_index[word]
        except KeyError:
            return self.append(word)

    def __contains__(self, word):
        self.append(word)
        return True


def merge(*vocabularies, keys=None, concat=lambda key, word: key + '_' + word, return_map=False):
    """Merge a list of Vocabulary into one

    Examples
    --------
    >>> import mangoes.vocabulary
    >>> v1 = mangoes.Vocabulary(['a', 'b', 'c'], language='l1')
    >>> v2 = mangoes.Vocabulary(['a', 'd'], language='l2')
    >>> merge(v1, v2)
    Vocabulary(['a', 'b', 'c', 'd'])
    >>> merge(v1, v2, keys=True)
    Vocabulary(['l1_a', 'l1_b', 'l1_c', 'l2_a', 'l2_d'])
    >>> merge(v1, v2, keys=['v1', 'v2'])
    Vocabulary(['v1_a', 'v1_b', 'v1_c', 'v2_a', 'v2_d'])

    With tokens :
    >>> import collections
    >>> Token = collections.namedtuple('Token', 'lemma POS')
    >>> v3 = mangoes.Vocabulary([Token('a', 'X'), Token('b', 'Y'), Token('c', 'X')], language='l1')
    >>> v4 = mangoes.Vocabulary([Token('a', 'X'), Token('d', 'Y')], language='l2')
    >>> LangToken = collections.namedtuple('LangToken', 'lemma POS lang')
    >>> merge(v3, v4, keys=True, concat=lambda lang, token: LangToken(*token, lang)) #doctest: +ELLIPSIS
    Vocabulary([LangToken(lemma='a', POS='X', lang='l1'), ..., LangToken(lemma='d', POS='Y', lang='l2')])

    Parameters
    ----------
    vocabularies: list of vocabularies
    keys: None (default), or bool or list of str
        If None or False, words that are common to several vocabularies are considered the same and will appear
        only once in resulting Vocabulary
        If keys is a list of string, of same size as counts, all words are prefixes with these keys.
        If keys is True, the languages of the vocabularies are used as keys
    concat: callable, optional
        Function that takes a key and a word (or a token) as input and returns a new word (or token)
        If keys are given, this function is called to create the word of the merged vocabulary from the given keys and
        the original words
        Default is '{key}_{word}' that prefixes each word with their key and is only valid from simple string words
        vocabularies.
        Bigrams are transformed applying these function to both of their part


    Returns
    -------
    Vocabulary

    """
    indices_map = []

    def _concat(k, w):
        if isinstance(w, mangoes.vocabulary.Bigram):
            return mangoes.vocabulary.Bigram(concat(k, w[0]), concat(k, w[1]))
        return concat(k, w)

    if not keys:
        merged = Vocabulary([])
        for v in vocabularies:
            _, m = merged.extend(v, inplace=True, return_map=True)
            if return_map:
                indices_map.append(m)
        if return_map:
            return merged, indices_map
        return merged
    else:
        if keys == True:
            keys = [v.language for v in vocabularies]
        merged = mangoes.Vocabulary([_concat(k, w) for v, k in zip(vocabularies, keys) for w in v])
        merged._params = {k: v._params for k, v in zip(keys, vocabularies)}
        if return_map:
            for v, k in zip(vocabularies, keys):
                indices_map.append({i: merged.index(_concat(k, w))
                                    for w, i in v.word_index.items()})
            return merged, indices_map
    return merged


def create_token_filter(fields):
    """Returns a function to filter the given fields from a token

    Examples
    --------
    >>> Token = mangoes.corpus.BROWN.Token
    >>> cat_token = Token(form="cat", lemma="cat", POS="NOUN")
    >>> mangoes.vocabulary.create_token_filter("lemma")(cat_token)
    'cat'
    >>> mangoes.vocabulary.create_token_filter(("lemma", "POS"))(cat_token)
    Token(lemma='cat', POS='NOUN')

    Parameters
    ----------
    fields: str or tuple
        name of the fields(s) to keep

    Returns
    -------
    callable


    """
    if fields:
        if isinstance(fields, str):
            def filter_token(token):
                if isinstance(token, Bigram):
                    return Bigram(getattr(token[0], fields), getattr(token[1], fields))
                else:
                    return getattr(token, fields)
        else:
            Token = collections.namedtuple('Token', fields)

            def filter_token(token):
                if isinstance(token, Bigram):
                    return Bigram(Token(*[getattr(token[0], attr) for attr in fields]),
                                  Token(*[getattr(token[1], attr) for attr in fields]))
                else:
                    return Token(*[getattr(token, attr) for attr in fields])
    else:
        def filter_token(token):
            return token
    return filter_token


def create_tokens_filter(fields):
    """Returns a function to filter the given fields from a list of tokens"""
    if fields:
        filter_token = create_token_filter(fields)

        def filter_tokens(sentence):
            return [filter_token(token) for token in sentence]
    else:
        def filter_tokens(sentence):
            return sentence

    return filter_tokens


def create_bigrams_filter(bigrams=None):
    """Returns a function to find expected bigrams within a sentence"""
    if bigrams:
        if getattr(bigrams[0][0], '_fields', None):
            # bigrams of Tokens
            fields = bigrams[0][0]._fields
            Token = collections.namedtuple('Token', fields)

            def filter_bigrams(sentence):
                filtered_sentence = []
                i = 0
                while i < len(sentence):
                    try:
                        x = Token(*[getattr(sentence[i], f) for f in fields]), Token(*[getattr(sentence[i + 1], f) for f in fields])
                        if x in bigrams:
                            filtered_sentence.append(Bigram(sentence[i], sentence[i + 1]))
                            i += 1
                        else:
                            filtered_sentence.append(sentence[i])
                    except IndexError:
                        # sentence[i] or sentence[i + 1] is None
                        filtered_sentence.append(sentence[i])
                    i += 1
                return filtered_sentence
        else:
            # bigrams of strings
            def filter_bigrams(sentence):
                filtered_sentence = []
                i = 0
                while i < len(sentence):
                    try:
                        if (sentence[i], sentence[i + 1]) in bigrams:
                            filtered_sentence.append(Bigram(sentence[i], sentence[i + 1]))
                            i += 1
                        else:
                            filtered_sentence.append(sentence[i])
                    except (TypeError, IndexError):
                        # sentence[i] or sentence[i + 1] is None
                        filtered_sentence.append(sentence[i])
                    i += 1
                return filtered_sentence
    else:
        def filter_bigrams(sentence):
            return sentence

    return filter_bigrams



# -*- coding: utf-8 -*-
"""
This module is for preprocessing text and training subword tokenizer for use in BERT like models in the
mangoes.modeling.bert modules.
It provides an interface into the huggingface tokenizers library.
"""
import os
import fileinput
import json
import warnings

import tokenizers

from mangoes import Vocabulary


class MixinToVocab:
    """
    Mixin class for creating Vocabulary class from tokenizer
    """
    def __init__(self):
        return

    def make_vocab_object(self):
        """
        Returns mangoes.Vocabulary object from tokenizer's vocab
        """
        return Vocabulary(self.get_vocab())


class CustomTokenizer(MixinToVocab, tokenizers.implementations.BaseTokenizer):
    """
    Class for using custom built tokenizer.
    Use this class if the other classes in this module do not contain the tokenizer you would like to use.
    This class inherits from tokenizers.BaseTokenizer() (and as such, tokenizers.Tokenizer()). For more
    information on available methods, see:
    https://huggingface.co/docs/tokenizers/python/latest/api/reference.html#tokenizer
    and
    https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/implementations/base_tokenizer.py

    Parameters
    ----------
    tokenizer: tokenizer.Tokenizer()
    """
    def __init__(self, tokenizer):
        tokenizers.implementations.BaseTokenizer.__init__(self, tokenizer)
        MixinToVocab.__init__(self)

    @staticmethod
    def from_file(path):
        tokenizer = tokenizers.Tokenizer.from_file(path)
        return CustomTokenizer(tokenizer)

    def train(self, trainer, files):
        """
        Train the tokenizer using a custom built Trainer

        Parameters
        ----------
        trainer: tokenizers.Trainer().
            Instantiated tokenizers.Trainer(). Make sure the trainer matches the tokenization model in the tokenizer
            attribute.
        files: file or list of files to train on.
        """
        if isinstance(files, str):
            files = [files]
        self._tokenizer.train(trainer, files)

    def save(self, path, pretty=False):
        """
        Save serialized tokenizer. Alternative to `save_model(path)` that includes attribute/keywords in saved file.

        Parameters
        ----------
        path: str
            Path to saved Tokenizer file. Can be directory or filename. If directory, tokenizer will be saved in
            path/tokenizer.json. Due to how transformers loads tokenizers, filename must be "tokenizer.json" if to be
            loaded by BERT model in mangoes.modeling.bert module.
        pretty: boolean
            Whether the JSON file should be pretty formatted.

        Returns
        --------
        path: str
            path where tokenizer is saved
        """
        if os.path.isdir(path):
            path = os.path.join(path, "tokenizer.json")
        elif not os.path.split(path)[1] == "tokenizer.json":
            warnings.warn("Tokenizer must be saved as `tokenizer.json` to be loaded by BERT model", RuntimeWarning)
        tokenizers.implementations.BaseTokenizer.save(self, path, pretty)
        return path


class WordLevelTokenizer(MixinToVocab, tokenizers.implementations.BaseTokenizer):
    """
    Simple word level tokenization.
    This class inherits from tokenizers.BaseTokenizer() (and as such, tokenizers.Tokenizer()). For more
    information on available methods, see:
    https://huggingface.co/docs/tokenizers/python/latest/api/reference.html#tokenizer
    and
    https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/implementations/base_tokenizer.py

    Parameters
    ----------
    vocab: str
        path to saved vocab or dict of [str -> int], if vocab is precomputed
    unk_token: str
        token to use for unknown tokens, defaults to "<unk>"
    unicode_normalizer: str
        optional, one of "nfc", "nfd", "nfkc", "nfkd", or None
    lowercase: Boolean
        whether or not to lowercase all text as part of tokenization pipeline
    """
    def __init__(self, vocab=None, unk_token="<unk>", unicode_normalizer=None, lowercase=False):
        if not vocab:
            index = {unk_token: 0}
        elif isinstance(vocab, dict):
            if unk_token not in vocab:
                vocab[unk_token] = len(vocab)
            index = vocab
        else:
            # have to check if unknown token in vocab file, else causes error
            vocab_file = vocab
            with open(vocab_file, "r") as filename:
                index = json.load(filename)
            if unk_token not in index:
                index[unk_token] = len(index)

        tokenizer = tokenizers.Tokenizer(tokenizers.models.WordLevel(index, unk_token=unk_token))
        normalizers = []
        if unicode_normalizer:
            normalizers += [tokenizers.normalizers.unicode_normalizer_from_str(unicode_normalizer)]
        if lowercase:
            normalizers += [tokenizers.normalizers.Lowercase()]
        if len(normalizers) > 0:
            if len(normalizers) > 1:
                tokenizer.normalizer = tokenizers.normalizers.Sequence(normalizers)
            else:
                tokenizer.normalizer = normalizers[0]

        tokenizer.pre_tokenizer = tokenizers.pre_tokenizers.WhitespaceSplit()
        parameters = {
            "model": "WordLevel",
            "lowercase": lowercase,
            "unicode_normalizer": unicode_normalizer,
            "unk_token": unk_token,
        }
        tokenizers.implementations.BaseTokenizer.__init__(self, tokenizer, parameters)
        MixinToVocab.__init__(self)

    def save(self, path, pretty=False):
        """
        Save serialized tokenizer. Alternative to `save_model(path)` that includes attribute/keywords in saved file.

        Parameters
        ----------
        path: str
            Path to saved Tokenizer file. Can be directory or filename. If directory, tokenizer will be saved in
            path/tokenizer.json. Due to how transformers loads tokenizers, filename must be "tokenizer.json" if to be
            loaded by BERT model in mangoes.modeling.bert module.
        pretty: boolean
            Whether the JSON file should be pretty formatted.

        Returns
        --------
        path: str
            path where tokenizer is saved
        """
        if os.path.isdir(path):
            path = os.path.join(path, "tokenizer.json")
        elif not os.path.split(path)[1] == "tokenizer.json":
            warnings.warn("Tokenizer must be saved as `tokenizer.json` to be loaded by BERT model", RuntimeWarning)
        tokenizers.implementations.BaseTokenizer.save(self, path, pretty)
        return path

    @staticmethod
    def from_file(vocab_filename, **kwargs):
        """
        Load a tokenizer from a saved vocab file

        Parameters
        ----------
        vocab_filename: str
            Path to saved vocab file.
        kwargs

        Returns
        -------
        WordLevelTokenizer object
        """
        return WordLevelTokenizer(vocab_filename, **kwargs)

    def train(self, files):
        """
        Train the model using the given file or path to directory containing files

        Parameters
        ----------
        files: str or List[str]
            path or paths to files to train tokenizer on.
        """
        def hook_compressed_encoded_text(filename, mode='r', encoding='utf-8'):
            ext = os.path.splitext(filename)[1]
            if ext == '.gz':
                import gzip
                return gzip.open(filename, mode + 't', encoding=encoding)
            return open(filename, mode, encoding=encoding)

        if isinstance(files, str):
            files = [files]

        with fileinput.input(files, mode="r", openhook=hook_compressed_encoded_text) as input_file:
            for line in input_file:
                if self._tokenizer.normalizer:
                    line = self._tokenizer.normalizer.normalize_str(line)
                words = line.split()
                self._tokenizer.add_tokens(words)


class ByteBPETokenizer(MixinToVocab, tokenizers.ByteLevelBPETokenizer):
    """
    Byte-level BPE, as used by Open-AI's GPT2 model
    This class inherits from tokenizers.ByteLevelBPETokenizer() (and as such, tokenizers.Tokenizer()). For more
    information on available methods, see:
    https://huggingface.co/docs/tokenizers/python/latest/api/reference.html#tokenizer
    and
    https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/implementations/byte_level_bpe.py

    Parameters
    ----------
    vocab: str or Dict[str, int]]
        path to saved vocab or dict of [str -> int], if vocab is precomputed
    merges: str or Dict[Tuple[int, int] or Tuple[int, int]]]
        path to saved merges or precomputed merges.
    add_prefix_space: Boolean
    lowercase: Boolean
        Whether or not to lowercase all text as part of tokenization pipeline
    dropout: Float
        The BPE dropout to use. Must be an float between 0 and 1
    unicode_normalizer: str
        Optional, one of "nfc", "nfd", "nfkc", "nfkd", or None
    continuing_subword_prefix: str
        The prefix to attach to subword units that don't represent a beginning of word.
    end_of_word_suffix: str
       The suffix to attach to subword units that represent an end of word.
    trim_offsets: Boolean
        Whether to trim the whitespaces from the produced offsets.
    """
    def __init__(self, vocab=None, merges=None, add_prefix_space=False, lowercase=False, dropout=None,
                 unicode_normalizer=None, continuing_subword_prefix=None, end_of_word_suffix=None, trim_offsets=False):
        tokenizers.ByteLevelBPETokenizer.__init__(self, vocab, merges, add_prefix_space, lowercase, dropout,
                                                  unicode_normalizer, continuing_subword_prefix, end_of_word_suffix,
                                                  trim_offsets)
        MixinToVocab.__init__(self)

    def save(self, path, pretty=False):
        """
        Save serialized tokenizer. Alternative to `save_model(path)` that includes attribute/keywords in saved file.

        Parameters
        ----------
        path: str
            Path to saved Tokenizer file. Can be directory or filename. If directory, tokenizer will be saved in
            path/tokenizer.json. Due to how transformers loads tokenizers, filename must be "tokenizer.json" if to be
            loaded by BERT model in mangoes.modeling.bert module.
        pretty: boolean
            Whether the JSON file should be pretty formatted.

        Returns
        --------
        path: str
            path where tokenizer is saved
        """
        if os.path.isdir(path):
            path = os.path.join(path, "tokenizer.json")
        elif not os.path.split(path)[1] == "tokenizer.json":
            warnings.warn("Tokenizer must be saved as `tokenizer.json` to be loaded by BERT model", RuntimeWarning)
        tokenizers.ByteLevelBPETokenizer.save(self, path, pretty)
        return path

    @staticmethod
    def from_file(vocab_filename, merges_filename, **kwargs):
        """
        Load a tokenizer from a saved vocab file

        Parameters
        ----------
        vocab_filename: str
            Path to saved vocab file.
        merges_filename: str
            Path to saved merges file.
        **kwargs

        Returns
        -------
        ByteBPETokenizer object
        """
        vocab, merges = tokenizers.models.BPE.read_file(vocab_filename, merges_filename)
        return ByteBPETokenizer(vocab, merges, **kwargs)


class CharacterBPETokenizer(MixinToVocab, tokenizers.CharBPETokenizer):
    """
    Original character level BPE tokenizer.
    This class inherits from tokenizers.CharBPETokenizer() (and as such, tokenizers.Tokenizer()). For more
    information on available methods, see:
    https://huggingface.co/docs/tokenizers/python/latest/api/reference.html#tokenizer
    and
    https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/implementations/char_level_bpe.py

    Parameters
    ----------
    vocab: str or Dict[str, int]]
        path to saved vocab or dict of [str -> int], if vocab is precomputed
    merges: str or Dict[Tuple[int, int] or Tuple[int, int]]]
        path to saved merges or precomputed merges.
    unk_token: str
        token to use for unknown tokens, defaults to "<unk>"
    suffix: str
       The suffix to attach to subword units that represent an end of word.
    dropout: Float
        The BPE dropout to use. Must be an float between 0 and 1
    lowercase: Boolean
        Whether or not to lowercase all text as part of tokenization pipeline
    unicode_normalizer: str
        Optional, one of "nfc", "nfd", "nfkc", "nfkd", or None
    bert_normalizer: Boolean
        Whether to use a bert normalizer, which cleans text and handles Chinese characters. Default is True.
    split_on_whitespace_only: Boolean
        Whether to split on whitespace only. Alternative if to split on whitespace and punctuation. Default is False.
    """
    def __init__(self, vocab=None, merges=None, unk_token="<unk>", suffix="</w>", dropout=None, lowercase=False,
                 unicode_normalizer=None, bert_normalizer=True, split_on_whitespace_only=False):
        tokenizers.CharBPETokenizer.__init__(self, vocab, merges, unk_token, suffix, dropout, lowercase,
                                             unicode_normalizer, bert_normalizer, split_on_whitespace_only)
        MixinToVocab.__init__(self)

    def save(self, path, pretty=False):
        """
        Save serialized tokenizer. Alternative to `save_model(path)` that includes attribute/keywords in saved file.

        Parameters
        ----------
        path: str
            Path to saved Tokenizer file. Can be directory or filename. If directory, tokenizer will be saved in
            path/tokenizer.json. Due to how transformers loads tokenizers, filename must be "tokenizer.json" if to be
            loaded by BERT model in mangoes.modeling.bert module.
        pretty: boolean
            Whether the JSON file should be pretty formatted.

        Returns
        --------
        path: str
            path where tokenizer is saved
        """
        if os.path.isdir(path):
            path = os.path.join(path, "tokenizer.json")
        elif not os.path.split(path)[1] == "tokenizer.json":
            warnings.warn("Tokenizer must be saved as `tokenizer.json` to be loaded by BERT model", RuntimeWarning)
        tokenizers.CharBPETokenizer.save(self, path, pretty)
        return path

    @staticmethod
    def from_file(vocab_filename, merges_filename, **kwargs):
        """
        Load a tokenizer from a saved vocab file

        Parameters
        ----------
        vocab_filename: str
            Path to saved vocab file.
        merges_filename: str
            Path to saved merges file.
        **kwargs

        Returns
        -------
        CharacterBPETokenizer object
        """
        vocab, merges = tokenizers.models.BPE.read_file(vocab_filename, merges_filename)
        return CharacterBPETokenizer(vocab, merges, **kwargs)


class BERTWordPieceTokenizer(MixinToVocab, tokenizers.BertWordPieceTokenizer):
    """
    Word piece tokenizer used for BERT.
    This class inherits from tokenizers.BertWordPieceTokenizer() (and as such, tokenizers.Tokenizer()). For more
    information on available methods, see:
    https://huggingface.co/docs/tokenizers/python/latest/api/reference.html#tokenizer
    and
    https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/implementations/bert_wordpiece.py

    Parameters
    ----------
    vocab: str or Dict[str, int]]
        path to saved vocab or dict of [str -> int], if vocab is precomputed
    unk_token: str
        token to use for unknown tokens, defaults to "<UNK>"
    sep_token: str
        token to use for separation, defaults to "<SEP>"
    cls_token: str
        token to use for class tokens, defaults to "<CLS>"
    pad_token: str
        token to use for padding, defaults to "<PAD>"
    mask_token: str
        token to use for masking, defaults to "<MASK>"
    handle_chinese_chars: Boolean
        Whether to handle chinese chars by putting spaces around them.
    strip_accents: Boolean
        Whether to strip all accents. If this option is not specified (ie == None), then it will be determined by the
        value for `lowercase` (as in the original Bert).
    lowercase: Boolean
        Whether or not to lowercase all text as part of tokenization pipeline.
    wordpieces_prefix: str
        The prefix to use for subwords that are not a beginning-of-word.
    """
    def __init__(self, vocab=None, unk_token="[UNK]", sep_token="[SEP]", cls_token="[CLS]", pad_token="[PAD]",
                 mask_token="[MASK]", clean_text=True, handle_chinese_chars=True, strip_accents=None, lowercase=True,
                 wordpieces_prefix="##"):
        tokenizers.BertWordPieceTokenizer.__init__(self, vocab, unk_token, sep_token, cls_token, pad_token, mask_token,
                                                   clean_text, handle_chinese_chars, strip_accents, lowercase,
                                                   wordpieces_prefix)
        MixinToVocab.__init__(self)

    def save(self, path, pretty=False):
        """
        Save serialized tokenizer. Alternative to `save_model(path)` that includes attribute/keywords in saved file.

        Parameters
        ----------
        path: str
            Path to saved Tokenizer file. Can be directory or filename. If directory, tokenizer will be saved in
            path/tokenizer.json. Due to how transformers loads tokenizers, filename must be "tokenizer.json" if to be
            loaded by BERT model in mangoes.modeling.bert module.
        pretty: boolean
            Whether the JSON file should be pretty formatted.

        Returns
        --------
        path: str
            path where tokenizer is saved
        """
        if os.path.isdir(path):
            path = os.path.join(path, "tokenizer.json")
        elif not os.path.split(path)[1] == "tokenizer.json":
            warnings.warn("Tokenizer must be saved as `tokenizer.json` to be loaded by BERT model", RuntimeWarning)
        tokenizers.BertWordPieceTokenizer.save(self, path, pretty)
        return path

    @staticmethod
    def from_file(vocab, **kwargs):
        """
        Load a tokenizer from a saved vocab file

        Parameters
        ----------
        vocab: str
            Path to saved vocab file.
        kwargs

        Returns
        -------
        BERTWordPieceTokenizer object
        """
        vocab = tokenizers.models.WordPiece.read_file(vocab)
        return BERTWordPieceTokenizer(vocab, **kwargs)


class SentPieceBPETokenizer(MixinToVocab, tokenizers.SentencePieceBPETokenizer):
    """
    BPE tokenizer with Sentence Piece pretokenization
    This class inherits from tokenizers.SentencePieceBPETokenizer() (and as such, tokenizers.Tokenizer()). For more
    information on available methods, see:
    https://huggingface.co/docs/tokenizers/python/latest/api/reference.html#tokenizer
    and
    https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/implementations/sentencepiece_bpe.py

    Parameters
    ----------
    vocab: str or Dict[str, int]]
        path to saved vocab or dict of [str -> int], if vocab is precomputed
    merges: str or Dict[Tuple[int, int] or Tuple[int, int]]]
        path to saved merges or precomputed merges.
    unk_token: str
        token to use for unknown tokens, defaults to "<unk>"
    replacement: str
        The replacement character for whitespace. Must be exactly one character. Default is the `▁` (U+2581) meta symbol
        (Same as in SentencePiece).
    add_prefix_space: Boolean
        Whether to add a space to the first word if there isn't already one. Default is True.
    dropout: Float
        The BPE dropout to use. Must be an float between 0 and 1
    """

    def __init__(self, vocab=None, merges=None, unk_token="<unk>", replacement="▁", add_prefix_space=True,
                 dropout=None):
        tokenizers.SentencePieceBPETokenizer.__init__(self, vocab, merges, unk_token, replacement, add_prefix_space,
                                                      dropout)
        MixinToVocab.__init__(self)

    def save(self, path, pretty=False):
        """
        Save serialized tokenizer. Alternative to `save_model(path)` that includes attribute/keywords in saved file.

        Parameters
        ----------
        path: str
            Path to saved Tokenizer file. Can be directory or filename. If directory, tokenizer will be saved in
            path/tokenizer.json. Due to how transformers loads tokenizers, filename must be "tokenizer.json" if to be
            loaded by BERT model in mangoes.modeling.bert module.
        pretty: boolean
            Whether the JSON file should be pretty formatted.

        Returns
        --------
        path: str
            path where tokenizer is saved
        """
        if os.path.isdir(path):
            path = os.path.join(path, "tokenizer.json")
        elif not os.path.split(path)[1] == "tokenizer.json":
            warnings.warn("Tokenizer must be saved as `tokenizer.json` to be loaded by BERT model", RuntimeWarning)
        tokenizers.SentencePieceBPETokenizer.save(self, path, pretty)
        return path

    @staticmethod
    def from_file(vocab_filename, merges_filename, **kwargs):
        """
        Load a tokenizer from a saved vocab file

        Parameters
        ----------
        vocab_filename: str
            Path to saved vocab file.
        merges_filename: str
            Path to saved merges file.
        **kwargs

        Returns
        -------
        SentPieceBPETokenizer object
        """
        vocab, merges = tokenizers.models.BPE.read_file(vocab_filename, merges_filename)
        return SentPieceBPETokenizer(vocab, merges, **kwargs)



from setuptools import setup, Extension


setup(
    name='mangoes',
    version='2.0.0',
    python_requires='>=3.6',
    packages=['mangoes', 'mangoes.evaluation', 'mangoes.utils', 'mangoes.modeling'],
    ext_modules=[Extension("mangoes.utils.counting", ["mangoes/utils/counting.pyx"])],
    package_data={
        'mangoes': ['resources/en/similarity/*.txt', 'resources/fr/similarity/*.txt', 'resources/en/analogy/*/*/*.txt',
                    'resources/en/outlier_detection/*/*.txt', 'resources/en/outlier_detection/*.zip'],
    },
    include_package_data=True,
    url='https://gitlab.inria.fr/magnet/mangoes/',
    download_url='https://gitlab.inria.fr/magnet/mangoes/repository/2.0.0/archive.tar.gz',
    license='LGPL',
    author='Inria - Magnet',
    author_email='nathalie.vauquier@inria.fr',
    description='Mangoes 2.0 is a toolbox for constructing and evaluating static or contextual token vector '
                'representations (aka word embeddings).',
    # numpy 1.20 does not work with python 3.6 and lower.
    # If mangoes becomes only python 3.7 and higher, we can get rid of the numpy version requirement
    install_requires=['cython', 'nltk', 'numpy<=1.19',
                      'scipy', 'sklearn', 'pandas', 'transformers>=4.2.2', 'torch>=1.7'],
    extras_require={
        'visualize': ["matplotlib"],
        'generator': ["gensim"]
    },
    classifiers=[  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Programming Language :: Python :: 3.6',
    ],

)



# -*-coding:utf-8 -*
"""Definitions a the context of a word in a sentence

This module defines classes to be used as `context` parameter in the
:func:`mangoes.counting.count_cooccurrence` function.

Examples
--------

    >>> window_3 = mangoes.context.Window(window_half_size=3)
    >>> contexts = window_3(['The', 'quick', 'brown', 'fox', 'jumps', 'over', 'the', 'lazy', 'dog'])
    >>> contexts[0]
    ['quick', 'brown', 'fox']
    >>> contexts[4]
    ['quick', 'brown', 'fox', 'over', 'the', 'lazy']
    >>> contexts[5]
    ['brown', 'fox', 'jumps', 'the', 'lazy', 'dog']
    >>> contexts[7]
    ['jumps', 'over', 'the', 'dog']

"""
import abc
import logging
import random
import warnings

from nltk.util import ngrams

import mangoes
import mangoes.utils.exceptions

_logger = logging.getLogger(__name__)


class Context:
    """Base callable class to define the context of a word in a sentence

    Parameters
    ----------
    vocabulary: mangoes.Vocabulary or list of string
        vocabulary of the words to consider in the context. Other words are ignored.

    See Also
    --------
    :func:`mangoes.counting.count_cooccurrence`
    """

    def __init__(self, vocabulary=None):
        self._params = {}
        self.vocabulary = vocabulary
        try:
            self.filter_sentence = mangoes.vocabulary.create_tokens_filter(vocabulary.entity)
        except AttributeError:
            # no vocabulary or vocabulary has no attribute entity
            def no_filter(sentence):
                return sentence

            self.filter_sentence = no_filter

    @abc.abstractmethod
    def __call__(self, sentence, mask=False):
        """Returns the elements in the defined context for each word in sentence

        This function extracts the list of elements found in the context of the elements of a sentence

        Parameters
        ----------
        sentence: list
            The sentence as a list of words or tokens
        mask: list of bool
            List of booleans that indicates the positions in the sentence that can be ignored (value=False) :
            empty contexts will be returned for these words

        Returns
        -------
        list of the same size than sentence
        """
        pass

    @property
    def params(self):
        """Parameters of the context"""
        return self._params


class Sentence(Context):
    """Implements sentence context

     This context extracts the list of tokens in a sentence, around each word, barring the
     word itself.

    """

    def __init__(self, vocabulary=None):
        super().__init__(vocabulary)

    def __call__(self, sentence, mask=False):
        filtered_sentence = self.filter_sentence(sentence)
        return [[w for w in filtered_sentence[0:position] + filtered_sentence[position + 1:len(filtered_sentence)]
                 if not self.vocabulary or w in self.vocabulary]
                if not mask or mask[position] else [] for position in range(len(filtered_sentence))]


class Window(Context):
    """Implements window-type context

    This context extracts the list of elements found in a window defined around each word, barring the
    word itself.

    Examples
    --------

    >>> import mangoes.context
    >>> sentence = 'Beautiful is better than ugly'.split()
    >>> Window()(sentence)
    [['is'], ['Beautiful', 'better'], ['is', 'than'], ['better', 'ugly'], ['than']]
    >>> Window(size=2)(sentence) # doctest: +NORMALIZE_WHITESPACE
    [['is', 'better'],
     ['Beautiful', 'better', 'than'],
     ['Beautiful', 'is', 'than', 'ugly'],
     ['is', 'better', 'ugly'],
     ['better', 'than']]
    >>> Window(size=3, n_grams=2)(sentence) # doctest: +NORMALIZE_WHITESPACE
    [[('is', 'better'), ('better', 'than')],
     [('better', 'than'), ('than', 'ugly')],
     [('Beautiful', 'is'), ('than', 'ugly')],
     [('Beautiful', 'is'), ('is', 'better')],
     [('is', 'better'), ('better', 'than')]]
    >>> Window(size=2, distance=True)(sentence) # doctest: +NORMALIZE_WHITESPACE
     [[('is', 1), ('better', 2)],
      [('Beautiful', -1), ('better', 1), ('than', 2)],
      [('Beautiful', -2), ('is', -1), ('than', 1), ('ugly', 2)],
      [('is', -2), ('better', -1), ('ugly', 1)],
      [('better', -2), ('than', -1)]]

    Parameters
    ----------
    size: int for symmetric window, tuple of 2 int if not
        size of the search space to the left and to the right of each word (default = 1).
        If an integer is given, the window will be centered around the word; asymmetric if a couple of int

    dirty: boolean (def = False)
        if True and some of the words in the window aren't correct ids (ex : -1), they will not be fetched, but the
        window will be extended further so as to still be able to meet the quota of 2*'window_half_size' (if
        symmetric) words to fetch.

    dynamic: boolean
        if True, the size of the actual window is sampled between 1 and `window_half_size`

    n_grams: int (default: 1)
        if n_grams > 1, an element of the list will be a n-grams of words instead of a single word.

    distance: boolean (default: False)
        if True, the distance from the target word will be added

    """

    def __init__(self, vocabulary=None, size=1, dirty=False, dynamic=False, n_grams=1, distance=False,
                 symmetric=None, window_half_size=None):
        super().__init__(vocabulary)

        if window_half_size or symmetric is not None:
            warnings.warn("'window_half_size' and 'symmetric' were replaced by the single parameter 'size' and"
                          "will be removed soon.", DeprecationWarning)
            # TODO: remove params in next release
            if size == 1:  # default
                size = window_half_size

        self._params.update({"size": size, "dirty": dirty,
                             "dynamic": dynamic, "n_grams": n_grams, "distance": distance})

        if isinstance(size, tuple):
            self.size_before, self.size_after = size
        else:
            self.size_before = self.size_after = size

        self._window = _create_window_method(self.size_before, self.size_after,
                                             dirty, dynamic, n_grams, distance)

    def __call__(self, sentence, mask=False):
        sentence = [w if not self.vocabulary or w in self.vocabulary else None for w in self.filter_sentence(sentence)]
        if not mask:
            mask = [True for _ in sentence]
        return self._window(sentence, mask)

    def call_on_encoded(self, encoded_sentence, mask=False):
        # TODO: find a better method name
        if not mask:
            mask = [True for _ in encoded_sentence]
        return self._window(encoded_sentence, mask)


def _create_window_method(size_before, size_after, dirty, dynamic, n, distance):
    """This function creates a "window" function according to the given parameters

    Returns
    -------
    _window : callable

    """
    return mangoes.utils.counting.get_window_function(size_before, size_after, dirty, dynamic, n, distance)


def _get_random_size(max_size_before, max_size_after):
    size_before = random.randint(1, max_size_before) if max_size_before > 0 else 0
    if max_size_after == max_size_before:
        size_after = size_before
    else:
        size_after = random.randint(1, max_size_after) if max_size_after > 0 else 0
    return size_before, size_after


class DependencyBasedContext(Context):
    """Implements Dependency-Based context

    Returns the modifiers and the head of each element of a sentence.

    Examples
    --------
    >>> source = mangoes.corpus.CONLLU(["1	australian	australian	ADJ	JJ	_	2	amod	_	_",
    >>>                                 "2	scientist	scientist	NOUN	NN	_	3	nsubj	_	_",
    >>>                                 "3	discovers	discover	VERB	VBZ	_	0	root	_	_",
    >>>                                 "4	star	star	NOUN	NN	_	3	dobj	_	_",
    >>>                                 "5	with	with	ADP	IN	_	6	case	_	_",
    >>>                                 "6	telescope	telescope	NOUN	NN	_	3	nmod	_	_"])
    >>> sentence = source.sentences().__next__()
    >>> context = mangoes.context.DependencyBasedContext(labels=True)
    >>> context(sentence)[1] # scientist
    {"australian/amod", "discovers/nsubj-"}

    References
    ----------
    .. [1] Levy, O., & Goldberg, Y. (2014, June). Dependency-Based Word Embeddings. In ACL (2) (pp. 302-308).

    Parameters
    ----------
    dependencies: {'universal-dependencies', 'stanford-dependencies'} or callable
        Representation used for dependencies annotation. Default is 'universal-dependencies'.
        You can also provide your own parser

    collapse: bool
        Whether or not the preposition relations should be collapsed. Default is False.

    labels: bool
        Whether or not the labels should be added to the output contexts. Default is False

    """
    def __init__(self, vocabulary=None, entity="form", dependencies="universal-dependencies",
                 collapse=False, labels=False, depth=1):
        super().__init__(vocabulary)

        if labels:
            self.filter_vocabulary = self.vocabulary
            self.vocabulary = mangoes.vocabulary.DynamicVocabulary()
            self._format_context = self._format_context_with_label
        else:
            self.filter_vocabulary = self.vocabulary
            self._format_context = self._format_context_without_label

        self.entity = entity
        self.filter_token = mangoes.vocabulary.create_token_filter(entity)

        self._params.update({"parser": dependencies, "collapse": collapse, "labels": labels, "depth": depth})
        if dependencies == "universal-dependencies":
            self.sentence_parser = self.ud_sentence_parser
        elif dependencies == "stanford-dependencies":
            self.sentence_parser = self.stanford_dependencies_sentence_parser
        elif callable(dependencies):
            self.sentence_parser = dependencies
        else:
            raise mangoes.utils.exceptions.NotAllowedValue(dependencies, ["universal-dependencies", "stanford-dependencies"])

    @property
    def collapse(self):
        return self._params["collapse"]

    @property
    def labels(self):
        return self._params["labels"]

    @property
    def depth(self):
        return self._params["depth"]

    def __call__(self, sentence, mask=False):
        dependency_tree = self.sentence_parser(sentence, self.collapse)

        for _ in range(self.depth - 1):
            dependency_tree = self.add_children(dependency_tree)

        contexts = [set() for _ in sentence]
        for i, token_dependencies in enumerate(dependency_tree):
            if token_dependencies:
                head = self.filter_token(sentence[i])
                for target_id, label in token_dependencies:
                    target = self.filter_token(sentence[target_id])
                    if not self.filter_vocabulary or target in self.filter_vocabulary:
                        contexts[i].add(self._format_context(target, label, False))
                    if not self.filter_vocabulary or head in self.filter_vocabulary:
                        contexts[target_id].add(self._format_context(head, label, True))

        return contexts

    def _format_context_with_label(self, token, label, inverse):
        return token + "/" + label + "-"*inverse

    def _format_context_without_label(self, token, label, inverse):
        return token

    @staticmethod
    def add_children(sentence_tree):
        new_sentence_tree = []
        for token_children in sentence_tree:
            new_children = set(token_children)
            for child, child_label in token_children:
                if sentence_tree[child]:
                    for grand_child, grand_child_label in sentence_tree[child]:
                        new_children.add((grand_child, child_label + '_' + grand_child_label))
            new_sentence_tree.append(new_children)
        return new_sentence_tree

    @staticmethod
    def ud_sentence_parser(sentence, collapse=False):
        """Returns an adjacency list from a sentence annotated with Universal Dependencies

        Examples
        --------
        >>> source = mangoes.corpus.CONLLU(["1	australian	australian	ADJ	JJ	_	2	amod	_	_",
        >>>                                 "2	scientist	scientist	NOUN	NN	_	3	nsubj	_	_",
        >>>                                 "3	discovers	discover	VERB	VBZ	_	0	root	_	_",
        >>>                                 "4	star	star	NOUN	NN	_	3	dobj	_	_",
        >>>                                 "5	with	with	ADP	IN	_	6	case	_	_",
        >>>                                 "6	telescope	telescope	NOUN	NN	_	3	nmod	_	_"])
        >>> sentence = source.sentences().__next__()
        >>> mangoes.context.DependencyBasedContext.ud_sentence_parser(sentence)
        [set(),
         {(0, 'amod')},
         {(1, 'nsubj'), (3, 'dobj'), (5, 'nmod')},
         set(),
         set(),
         {(4, 'case')}]

        Parameters
        ----------
        sentence: list of Tokens

        collapse: boolean
            Whether or not to collapse prepositions

        Returns
        -------
        list of same size as sentence
            Returns the dependents of each token in the sentence.

        """
        root_label = "root"
        preposition_label = "case"
        preposition_object_label = "nmod"

        relations = [set() for _ in sentence]

        for token in sentence:
            try:
                if token.dependency_relation == root_label:
                    continue
                elif collapse and token.dependency_relation == preposition_label:
                    preposition_object_position = int(token.head) - 1
                    preposition_object = sentence[preposition_object_position]

                    head_position = int(preposition_object.head) - 1

                    relations[head_position].add(
                        (preposition_object_position, preposition_label + "_" + token.form))
                elif collapse and token.dependency_relation == preposition_object_label:
                    continue  # done in the previous case
                else:
                    token_position = int(token.id) - 1
                    head_position = int(token.head) - 1

                    relations[head_position].add((token_position, token.dependency_relation))
            except ValueError:
                logging.warning("Unable to parse token %s in sentence %s (ValueError)",
                                token, " ".join([t.form for t in sentence]))
                continue
            except IndexError:
                logging.warning("Unable to parse token %s in sentence %s (IndexError)",
                                token, " ".join([t.form for t in sentence]))
                continue

        return relations

    @staticmethod
    def stanford_dependencies_sentence_parser(sentence, collapse=False):
        """Returns an adjacency list from a sentence annotated with Stanford Dependencies

        Parameters
        ----------
        sentence: list of Tokens

        collapse: boolean
            Whether or not to collapse prepositions

        Returns
        -------
        list of same size as sentence
            Returns the dependents of each token in the sentence.

        """
        root_label = "root"
        preposition_label = "prep"
        preposition_object_label = "pobj"

        relations = [set() for _ in sentence]

        for token in sentence:
            if token.dependency_relation == root_label:
                continue
            elif collapse and token.dependency_relation == preposition_label:
                continue
            elif collapse and token.dependency_relation == preposition_object_label:
                preposition = sentence[(int(token.head) - 1)]
                token_position = int(token.id) - 1
                head_position = int(preposition.head) - 1

                relations[head_position].add((token_position, preposition_label + "_" + preposition.form))
            else:
                token_position = int(token.id) - 1
                head_position = int(token.head) - 1

                relations[head_position].add((token_position, token.dependency_relation))

        return relations



# -*- coding: utf-8 -*-
"""
This module provides an interface into the transformers BERT models, including pretrained models, pretraining from
scratch, and fine tuning.
The main BERT classes inherit from transformers BERT models and include tokenizers, allowing for easy inference and
feature extraction.
"""
import warnings
from abc import ABC, abstractmethod

import transformers
import numpy as np
import torch


def merge_subword_embeddings(embeddings, text, token_offsets, pretokenized=False):
    """
    Function to merge possible subword embeddings for a text to word embeddings, and filter out special token
    embeddings.
    If a word in the text is represented by two or more sub-words, this function will average the sub-word
    embeddings to create a single embedding vector for this word.
    Accepts torch tensors or numpy arrays.

    Parameters
    ----------
    embeddings: numpy array or torch tensor of shape (num_tokens, embedding_size) if one sentence, or (num_sentences,
        num_tokens, embedding_size) if multiple sentences
        the subword embeddings to be merged
    text: str or List[str] if pre_tokenized=False, else List[str] or List[List[str]]
        the text corresponding to the embedding output
    token_offsets: (char_start, char_end) for each token. Indices correspond to either text (in not pretokenized)
        or word (if pretokenized)
    pretokenized: if input into tokenizer was split into words

    Returns
    -------
    numpy array or torch tensor with sub-word embeddings averaged to form word embeddings, padded to number of words in
        longest sentence. Padded words will contain vectors of zeros.
    """
    if not pretokenized:
        words = [text.split()] if isinstance(text, str) else [sent.split() for sent in text]
    else:
        words = [text] if isinstance(text[0], str) else text
    embedding_size = embeddings.shape[-1]
    num_sentences = len(words)
    max_words = max([len(sent) for sent in words])
    if isinstance(embeddings, torch.Tensor):
        to_torch = True
        embeddings = embeddings.cpu().numpy()
    else:
        to_torch = False
    output_array = np.zeros((num_sentences, max_words, embedding_size), dtype=embeddings.dtype)
    for i in range(num_sentences):
        current_token_index = 0
        while token_offsets[i][current_token_index][1] == 0:
            # loop past any special tokens at the start
            current_token_index += 1
        for j in range(len(words[i])):
            current_start = current_token_index
            current_end = current_token_index
            while current_end < len(token_offsets[i]) - 1 and \
                    token_offsets[i][current_end][1] == token_offsets[i][current_end + 1][0]:
                current_end += 1
            if current_start == current_end:
                output_array[i, j, :] = embeddings[i, current_start, :]
            else:
                output_array[i, j, :] = np.mean(embeddings[i][current_start:current_end + 1][:], 0)
            current_token_index = current_end + 1
    if to_torch:
        output_array = torch.from_numpy(output_array)
    return output_array


class PipelineMixin(ABC):
    """
    Mixin class for tasks covered by transformers Pipelines.
    """
    def __init__(self):
        self.pipeline = None

    @abstractmethod
    def _construct_pipeline(self):
        """
        abstract method for creating inference pipeline
        """

    def _predict(self, inputs, **keyword_args):
        if not self.pipeline:
            self._construct_pipeline()
        if inputs:
            return self.pipeline(inputs, **keyword_args)
        return self.pipeline(**keyword_args)


class TransformerModel(ABC):
    """Base class for mangoes BERT models.
    Includes functionality for extracting embeddings from trained base Bert model.

    Parameters
    ----------
    device: int, or None
        if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
    """
    def __init__(self, device=None):
        self.model = None
        self.tokenizer = None
        if not device:
            # will use GPU for feature extraction, if available and no device input argument
            if torch.cuda.is_available():
                self.model_device = torch.device("cuda:0")
            else:
                self.model_device = torch.device("cpu")
        else:
            self.model_device = torch.device("cpu" if device < 0 else "cuda:{}".format(device))
            if self.model_device.type == "cuda":
                if torch.cuda.device_count() <= device:
                    warnings.warn(f"CUDA device {device} is not available on this machine, using CPU for BERT model",
                                  RuntimeWarning)
                    self.model_device = torch.device("cpu")

    @abstractmethod
    def train(self, output_dir=None, train_text=None, eval_text=None, collator=None, train_dataset=None,
              eval_dataset=None, trainer=None, **training_args):
        """
        abstract method for BERT training
        """

    @classmethod
    @abstractmethod
    def load(cls, pretrained_tokenizer, pretrained_model):
        """
        abstract method for loading a pretrained tokenizer and model
        """

    def save(self, output_directory, save_tokenizer=False):
        """
        Method to save BERT model and optionally save tokenizer. The tokenizer is already saved (the input to this class
        includes a pretrained tokenizer), but this method will save the tokenizer as well if needed.
        Both the tokenizer files and model files will be saved to the output directory. The output directory can be
        inputted as an argument to the "load()" method of the inheriting classes (for the model and tokenizer
        arguments)

        Parameters
        ----------
        output_directory: str
            path to directory to save model
        save_tokenizer: Boolean
            whether to save tokenizer in directory or not, defaults to False
        """
        self.model.save_pretrained(output_directory)
        if save_tokenizer:
            self.tokenizer.save_pretrained(output_directory)

    def generate_outputs(self, text, pre_tokenized=False, output_attentions=False, output_hidden_states=False,
                         word_embeddings=False):
        """
        Tokenize input text and pass it through the BERT model, optionally outputting hidden states or attention
        matrices.

        Parameters
        ----------
        text: str or List[str] if pre_tokenized=False, else List[str] or List[List[str]]
            the text to compute features for.
        pre_tokenized: Boolean
            whether or not the input text is pre-tokenized (ie, split on spaces)
        output_attentions: Boolean, optional, defaults to False
            Whether or not to return the attentions tensors of all attention layers.
        output_hidden_states: Boolean, optional, defaults to False
            Whether or not to return the hidden states of all layers.
        word_embeddings: Boolean
            whether or not to filter special token embeddings and average sub-word embeddings (hidden states) into word
            embeddings. If pre-tokenized inputs, the sub-word embeddings will be averaged into the tokens pass as i
            nputs.
            If pre-tokenized=False, the text will be split on whitespace and the sub-word embeddings will be averaged
            back into these words produced by splitting the text on whitespace.
            Only used if output_hidden_states = True.
            If False, number of output embeddings could be greater than (number of words + special tokens).
            If True, number of output embeddings == number of words, sub-words are averaged together to create word
            level embeddings and special token embeddings are excluded.

        Returns
        -------
        Dict containing (note that if single text sequence is passed as input, batch size will be 1):
        hidden_states: (Tuple (one for each layer) of torch.FloatTensor (batch_size, sequence_length, hidden_size)).
            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
            Only returned if output_hidden_states is True. If word_embeddings, the sequence length will be the number of
            words in the longest sentence, ie the maximum number of words. Shorter sequences will be padded with zeros.
        attentions: Tuple of torch.FloatTensor (one for each layer) of shape (batch_size, num_heads, sequence_length,
            sequence_length).
            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention
            heads.
            Only returned if output_attentions is True
        offset_mappings: Tensor of shape (batch_size, sequence_length, 2)
            Tensor containing (char_start, char_end) for each token, giving index into input strings of start and end
            character for each token. If input is pre-tokenized, start and end index maps into associated word. Note
            that special tokens are included with 0 for start and end indices, as these don't map into input text
            because they are added inside the function.
            This output is only available to tokenizers that inherit from transformers.PreTrainedTokenizerFast . This
            includes the BERT tokenizer and most other common tokenizers, but not all possible tokenizers in the
            library. If the tokenizer did not inherit from this class, this output value will be None.
        if PretrainedTransformerModelForFeatureExtraction:
            last_hidden_state: (torch.FloatTensor of shape (batch_size, sequence_length, hidden_size))
                Sequence of hidden-states at the output of the last layer of the model.
            pooler_output: (torch.FloatTensor of shape (batch_size, hidden_size))
                Last layer hidden-state of the first token of the sequence (classification token) further processed by a
                Linear layer and a Tanh activation function.
        if BERTForPreTraining:
            prediction_logits: (torch.FloatTensor of shape (batch_size, sequence_length, config.vocab_size))
                Prediction scores of the language modeling head (scores for each vocabulary token before SoftMax).
            seq_relationship_logits: (torch.FloatTensor of shape (batch_size, 2))
                Prediction scores of the next sequence prediction (classification) head (scores of True/False
                continuation before SoftMax).
        if BERTForMaskedLanguageModeling:
            logits: (torch.FloatTensor of shape (batch_size, sequence_length, config.vocab_size))
                Prediction scores of the language modeling head (scores for each vocabulary token before SoftMax).
        if BERTForSequenceClassification:
            logits: (torch.FloatTensor of shape (batch_size, config.num_labels))
                classification scores, before softmax
        if BERTForTokenClassification:
            logits: (torch.FloatTensor of shape (batch_size, sequence_length, config.num_labels))
                classification scores, before softmax
        """
        self.model.eval()
        inputs = self.tokenizer(text, is_split_into_words=pre_tokenized,
                                return_offsets_mapping=isinstance(self.tokenizer, transformers.PreTrainedTokenizerFast),
                                padding=True, return_tensors='pt')
        if isinstance(self.tokenizer, transformers.PreTrainedTokenizerFast):
            offset_mappings = inputs.pop('offset_mapping')
        else:
            offset_mappings = None
        with torch.no_grad():
            if self.model_device.type == "cuda":
                inputs = {name: tensor.to(self.model_device) for name, tensor in inputs.items()}
                if not next(self.model.parameters()).is_cuda:
                    self.model.to(self.model_device)
            results = self.model.forward(**inputs, return_dict=True,  output_attentions=output_attentions,
                                         output_hidden_states=output_hidden_states)
        if output_hidden_states and word_embeddings:
            if offset_mappings is not None:
                hidden_states = []  # tuples are immutable so need to create new and replace in results dict
                for i in range(len(results["hidden_states"])):
                    hidden_states.append(merge_subword_embeddings(results["hidden_states"][i], text,
                                                                  token_offsets=offset_mappings.numpy(),
                                                                  pretokenized=pre_tokenized))
                results["hidden_states"] = hidden_states
            else:
                warnings.warn("Tokenizer type does not support offset mappings, so word embedding consolidation is not "
                              "possible", RuntimeWarning)
        results["offset_mappings"] = offset_mappings
        return dict(results)


class PretrainedTransformerModelForFeatureExtraction(TransformerModel, PipelineMixin):
    """Class for using a pretrained transformer model and tokenizer. This class is meant to be used if you want to use a
    model that is not covered by the BERT classes in this module. It utilizes the AutoModel and AutoTokenizer classes
    in transformers. As it uses AutoModel (and not AutoModelForQuestionAnswering, for example), it can only be used
    for base pretrained models, and not fine tuning tasks.

    Parameters
    ----------
    pretrained_tokenizer: str
        Either:
            - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
              ``bert-base-uncased``.
            - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
              ``dbmdz/bert-base-german-cased``.
            - A path to a `directory` containing model weights saved using
              :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
    pretrained_model: str
        Either:
            - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
              ``bert-base-uncased``.
            - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
              ``dbmdz/bert-base-german-cased``.
            - A path to a `directory` containing model weights saved using
              :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
    device: int, optional, defaults to None
        if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available

    """
    def __init__(self, pretrained_tokenizer, pretrained_model, device=None):
        TransformerModel.__init__(self, device)
        PipelineMixin.__init__(self)
        self.tokenizer = transformers.AutoTokenizer.from_pretrained(pretrained_tokenizer)
        self.model = transformers.AutoModel.from_pretrained(pretrained_model)
        self.model.eval()

    def _construct_pipeline(self):
        """
        Implementation for creating inference pipeline.
        """
        self.model.eval()
        self.pipeline = transformers.FeatureExtractionPipeline(self.model, self.tokenizer,
                                                               device=-1 if self.model.device.type == "cpu" else
                                                               self.model.device.index)

    def predict(self, inputs, **kwargs):
        """
        Run input text through the feature extraction pipeline, extracting the hidden states of each layer.

        Parameters
        ----------
        inputs: str or list of strs
            inputs to extract features

        Returns
        -------
        nested list of float, hidden states.
        """
        return self._predict(inputs, **kwargs)

    @classmethod
    def load(cls, pretrained_tokenizer, pretrained_model, device=None):
        """
        Load a model and tokenizer. For this class, this method is essentially the same as the init function, and the
        two can be used interchangeably.

        Parameters
        ----------
        pretrained_tokenizer: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        pretrained_model: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        device: int, optional, defaults to None
            if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
        """
        model_object = cls(pretrained_tokenizer, pretrained_model, device)
        return model_object

    def train(self, output_dir=None, train_dataset=None, eval_dataset=None, collator=None, trainer=None,
              **training_args):
        """
        This function does nothing, use a task specific BERT class to pre-train or fine-tune
        """
        warnings.warn("This base model class does not implement training, it is meant for feature extraction")



# -*- coding: utf-8 -*-
"""Functions to visualize embeddings or some properties of Embeddings.

"""
import logging

import numpy as np
import matplotlib.pyplot as plt

import mangoes.evaluation.statistics

logger = logging.getLogger(__name__)


# ########################################
# t-SNE
def tsne(embeddings):
    """Create a 2d projections of the embeddings using t-SNE

    Parameters
    ----------
    embeddings: an instance of Embeddings
        Instance of :class:`mangoes.Embeddings` to project
    """

    try:
        matrix = embeddings.matrix.toarray()
    except AttributeError:
        # already dense
        matrix = embeddings.matrix

    import sklearn.manifold
    model = sklearn.manifold.TSNE()
    return model.fit_transform(matrix)


def plot_tsne(embeddings, words_to_display=None):
    """Plot a 2d projection of the embeddings using t-SNE

    Parameters
    ----------
    embeddings: an instance of Embeddings
        Instance of :class:`mangoes.Embeddings` to project
    words_to_display: list of str, optional
        list of words to display. If None (default), the first 100 words are displayed

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> fig = plt.figure()
    >>> mangoes.visualize.plot_tsne(embedding)
    >>> plt.show()
    """
    emb_2d = tsne(embeddings)

    plt.scatter(emb_2d[:, 0], emb_2d[:, 1], linewidth=0)

    if not words_to_display:
        words_to_display = embeddings.words[:min(100, len(embeddings.words))]
    for word in words_to_display:
        index = embeddings.words.index(word)
        plt.annotate(word, (emb_2d[index, 0], emb_2d[index, 1]), xycoords='data',
                     xytext=(10, 0), textcoords='offset points',
                     size=15, va="center", color='black',
                     bbox=dict(boxstyle="round", alpha=0.3, fc="white", ec="none"))


def plot_distances(embeddings, ax, nb_intervals=8, title="Distances between words", scale="symlog"):
    """Plot an histogram of the distances between all the words of the Embeddings

    Notes
    -----
    Use projection='polar' in matplotlib to get a circular histogram

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> fig = plt.figure()
    >>> ax = plt.subplot(111, projection='polar')
    >>> mangoes.visualize.plot_distances(embeddings, ax)

    Parameters
    ----------
    embeddings : an instance of Embeddings
        Instance of :class:`mangoes.Embeddings`
    ax :
        A matplotlib axes object
    nb_intervals : int, optional
        number of intervals (default : 8)
    title : str, optional
        title to use for the plot
    scale : {‘symlog’, ‘linear’, ‘log’, ‘logit’}
        scaling of the y-axis (default : symlog)
    """
    theta = np.linspace(0.0, np.pi, nb_intervals + 1, endpoint=True)
    ax.set_title(title)
    radii = mangoes.evaluation.statistics.distances_histogram(embeddings, theta)
    width = (np.pi / nb_intervals)
    bars = ax.bar(theta[:-1] + width/2, radii, width=width, bottom=0.0)
    ax.set_yscale(scale)

    for tick in ax.yaxis.get_major_ticks():
        tick.label.set_fontsize('xx-small')
    for tick in ax.xaxis.get_major_ticks():
        tick.label.set_fontsize('xx-small')

    for r, bar in zip(theta, bars):
        bar.set_facecolor(plt.cm.viridis(r / np.pi))


def plot_isotropy(embeddings, ax, title="Partition function value"):
    """Plot an histogram of the repartition of the values of a partition function

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> fig = plt.figure()
    >>> ax = plt.subplot(111)
    >>> mangoes.visualize.plot_isotropy(embeddings, ax)

    Parameters
    ----------

    embeddings :
        Instance of :class:`mangoes.Embeddings`
    ax :
        A matplotlib axes object
    title : str, optional
        title to use for the plot
    """
    # TODO : works only for dense matrices
    concentration, mean_value, values = mangoes.evaluation.statistics.isotropy_from_partition_function(embeddings)

    ax.set_title(title)
    ax.hist(values / mean_value, bins=40, range=(0, 2))

    ax.axvline(0.9, linestyle='dashed')
    ax.axvline(1.1, linestyle='dashed')
    ax.text(1, ax.get_ylim()[1] * 0.95, "{:.2%}".format(concentration), horizontalalignment='center', va='center')



# -*- coding: utf-8 -*-
"""Base classes used by evaluation tasks in `evaluation` package.

"""
import abc
import os
import zipfile
from collections import namedtuple

import pkg_resources

import mangoes
from mangoes.constants import ENCODING


# ####################################################################################################################
# DATASETS

class BaseDataset:
    """Base abstract class to define datasets

    """
    def __init__(self, name, data):
        """

        Parameters
        ----------
        name: str
            The name of the dataset
        data: list or dict or str
            If list, the list contains the "questions" of the dataset. The format depends on the task.
            If dict, each key is a subset of the dataset; the value can be a list or another subset
            If str, the string should be the path to a file, a zip or a directory containing the dataset
        """
        self.name = name
        if isinstance(data, str):
            if not self._check_resource_exists(data):
                raise mangoes.utils.exceptions.ResourceNotFound(path=data)
            self._data = data
        else:
            def parse_data(content):
                if isinstance(content, list):
                    return [self.parse_question(l) for l in content]
                else:
                    return {subset : parse_data(content[subset]) for subset in content}

            self._data = parse_data(data)

    @classmethod
    @abc.abstractmethod
    def parse_question(self, question):
        """Parse a string into a namedtuple representing a question for the given task"""
        pass

    @classmethod
    def parse_file(cls, file_content):
        return [cls.parse_question(line) for line in file_content if line.strip()]

    @property
    def data(self):
        # lazy load the data from files
        if isinstance(self._data, str):
            self._data = self._read_from(self._data)
        return self._data

    def get_subset(self, path):
        data = self.data
        for subset in path.split('/'):
            data = data[subset]
        return data

    @classmethod
    def _read_from(cls, path):
        if pkg_resources.resource_exists(__name__, path):
            path = pkg_resources.resource_filename(__name__, path)

        if os.path.isdir(path):
            return cls._read_from_dir(path)
        elif zipfile.is_zipfile(path):
            return cls._read_from_zip(path)
        else:
            try:
                return cls._read_from_file(path)
            except FileNotFoundError:
                raise mangoes.utils.exceptions.ResourceNotFound(path=path)

    @classmethod
    def _read_from_file(cls, path):
        with open(path, encoding=ENCODING) as file_content:
            return cls.parse_file(file_content)

    @classmethod
    def _read_from_dir(cls, path):
        questions = {}
        for subset in [os.path.join(path, subpath) for subpath in os.listdir(path)]:
            subset_name = _get_name(subset)
            if not subset_name.startswith('.'):
                questions[subset_name] = cls._read_from(subset)
        return questions

    @classmethod
    def _read_from_zip(cls, path):
        questions = {}
        with zipfile.ZipFile(path, 'r') as zip_file:
            for filename in zip_file.namelist():
                content = cls.parse_file(zip_file.read(filename).decode().split('\n'))
                if content:
                    subset_name = mangoes.evaluation.base._get_name(filename)
                    questions[subset_name] = content
        return questions

    @staticmethod
    def _check_resource_exists(path):
        if pkg_resources.resource_exists(__name__, path):
            return True

        return os.path.exists(path)


class BaseEvaluator:
    """Base abstract class for all Evaluator

    An Evaluator is a class that, given a representation, will predict results for the task it is implementing.
    """
    @abc.abstractmethod
    def predict(self, question):
        pass


class PrintableReport:
    """Base abstract class to print Evaluation reports"""
    LINE_LENGTH = 96
    COL = 4

    def __init__(self, evaluation, keep_duplicates=True):
        self.evaluation = evaluation
        self.keep_duplicates = keep_duplicates

    def __str__(self):
        return self.to_string()

    def to_string(self, show_subsets=False, show_questions=False):
        string = "\n"
        string += self._print_header()
        for dataset in self.evaluation.datasets:
            string += self._print_subset(dataset, None, show_subsets, show_questions)
            string += '-' * self.LINE_LENGTH + "\n"
        return string

    def _print_header(self):
        string = ""
        for line in self.HEADER:
            string += "{:>{width}}".format("", width=self.LINE_LENGTH - sum(line[1]) * self.COL)
            for h, l in zip(*line):
                string += "{:>{width}}".format(h, width=l * self.COL)
            string += "\n"
        return string + "=" * self.LINE_LENGTH + "\n"

    def _print_predictions_header(self):
        string = ""
        for line in self.PREDICTION_HEADER:
            string = "{:{width}}".format("", width=(self.LINE_LENGTH - sum(line[1]) * self.COL))
            for label, length in zip(*line):
                string += "{:>{width}}".format(label, width=length * self.COL)
            string += "\n"
        return string

    def _print_subset(self, dataset, subset_path=None, show_subsets=False, show_questions=False, indent=0):
        full_subset_path = dataset.name + '/' + subset_path if subset_path else dataset.name
        data = dataset.get_subset(subset_path) if subset_path else dataset.data

        string = self._print_score_line(full_subset_path, indent)

        if isinstance(data, list):
            if show_questions and data:
                string += self._print_predictions(full_subset_path, indent)
        elif show_subsets or show_questions:
            string += "\n"
            for subset in sorted(data):
                string += self._print_subset(dataset, (subset_path + '/' if subset_path else '') + subset,
                                             show_subsets, show_questions, indent + 1)
            if not show_questions: string += "\n"

        return string

    def _print_predictions(self, subset_path, indent=0):
        questions = self.evaluation._questions_by_subset[subset_path].questions#['questions']
        if questions:
            if not self.keep_duplicates:
                # remove duplicates preserving order
                import collections
                questions = list(collections.OrderedDict.fromkeys(questions))

            string = "\n"
            string += self._print_predictions_header()

            for similarity in questions:
                string += self._print_prediction(similarity, indent)

            return string + "\n"
        return "\n"

    def _print_score_line(self, name, indent=None):
        score = self.evaluation.get_score(dataset=name, keep_duplicates=self.keep_duplicates)
        nb_total_questions_in_original = self.evaluation._questions_by_subset[name].nb_total#["nb_total"]
        nb_duplicates = self.evaluation._questions_by_subset[name].nb_duplicates#["nb_duplicates"]

        line = "{:<{width}}".format('    ' * indent + name.split('/')[-1], width=(self.LINE_LENGTH - 3 * 3 * self.COL))
        line += "{:>{width}}".format("{}/{}".format(score.nb, nb_total_questions_in_original), width=3 * self.COL)

        line += self._print_score(score)

        if nb_duplicates:
            prefix = 'including ' if self.keep_duplicates else '-'
            plural = 's' if nb_duplicates > 1 else ''
            line += "\n{:>{width}}".format('    ' * indent + "({}{} duplicate{})".format(prefix, nb_duplicates, plural),
                                           width=(self.LINE_LENGTH - 3 * 2 * self.COL))
        return line + "\n"

    @abc.abstractmethod
    def _print_score(self, score):
        pass


class BaseEvaluation:
    _Report = PrintableReport
    _FilteredSubset = namedtuple('PreparedSubset', 'questions nb_total nb_oov nb_duplicates')
    # A Filtered Subset is the result of the filtering of a dataset, according to a representation:
    #     questions = the questions of this subset that can be answered with the representation
    #     nb_total = total number of questions in the original dataset, including oov and duplicates
    #     nb_oov = number of ignored questions because some terms are not in the represented vocabulary
    #     nb_duplicates = number of duplicates after removing oov questions

    def __init__(self, evaluator, *datasets, lower=True, evaluator_kwargs={}):
        self.evaluator = evaluator
        self.evaluator_kwargs = evaluator_kwargs
        self.datasets = datasets
        self.lower = lower

        self.predictions, self._questions_by_subset = self._prepare()

    @abc.abstractmethod
    def _filter_list_of_questions(self, list_of_questions):
        pass

    def _prepare(self):
        """
        Prepare the dataset(s) : count and remove oov questions, count duplicates,
        make predictions for all remaining questions
        Returns
        -------

        """
        questions_by_subset = {}
        unique_questions = set()

        def _prepare_data(name, data):
            questions_by_subset[name] = {}
            if isinstance(data, list):
                filtered_data, unique_questions_in_data, unique_with_gold_in_data = self._filter_list_of_questions(data)

                questions_by_subset[name] = filtered_data
                unique_questions.update(unique_questions_in_data)

                return unique_with_gold_in_data
            else:
                questions_in_data = []  # all ordered considered questions in this dataset, with duplicates
                unique_with_gold_in_data = set() # keep track of questions to count duplicates
                nb_total_in_data, nb_oov_in_data, nb_duplicates_in_data = 0, 0, 0

                for subset in data:
                    unique_with_gold_in_subset = _prepare_data(name + '/' + subset, data[subset])
                    filtered_subset = questions_by_subset[name + '/' + subset]

                    # add the considered questions of this subset
                    questions_in_data.extend(filtered_subset.questions)

                    nb_total_in_data += filtered_subset.nb_total
                    nb_oov_in_data += filtered_subset.nb_oov

                    # count duplicates : the ones within the subset ...
                    nb_duplicates_in_data += filtered_subset.nb_duplicates
                    # ... and the ones between the subsets
                    nb_duplicates_in_data += sum([q in unique_with_gold_in_data for q in set(filtered_subset.questions)])

                    unique_with_gold_in_data.update(unique_with_gold_in_subset)

                questions_by_subset[name] = self._FilteredSubset(questions_in_data,
                                                                 nb_total_in_data, nb_oov_in_data, nb_duplicates_in_data)
                return unique_with_gold_in_data

        for dataset in self.datasets:
            _prepare_data(dataset.name, dataset.data)

        return self.evaluator.predict(unique_questions, **self.evaluator_kwargs), questions_by_subset

    def get_score(self, dataset=None, keep_duplicates=True):
        """Return the score(s) of the evauation

        Parameters
        ----------
        dataset: str or None
            If several datasets were given, name of the dataset to get the score.
        keep_duplicates: bool
            Whether or not the duplicates should be kept or ignored to get the score

        Returns
        -------
        tuple or dict
            If only one dataset, returns the Score for this dataset : the Score is a namedtuple that depends on the task
            If several datasets, return a dict with datasets names as keys and Score as values

        """
        if dataset:
            if keep_duplicates:
                questions = self._questions_by_subset[dataset].questions
            else:
                questions = set(self._questions_by_subset[dataset].questions)

            return self._Score(*self._score(questions),
                               len(questions))
        else:
            if len(self.datasets) == 1:
                return self.get_score(self.datasets[0].name, keep_duplicates=keep_duplicates)
            return {dataset.name: self.get_score(dataset.name, keep_duplicates=keep_duplicates)
                    for dataset in self.datasets}

    def get_report(self, keep_duplicates=True, show_subsets=False, show_questions=False):
        """Gets a PrintableReport for this evaluation

        Parameters
        ----------
        keep_duplicates: bool
            Whether or not the duplicates should be kept or ignored to get the scores
        show_subsets
            Whether or not to see the scores for all subsets in the report
        show_questions
            Whether or not to see predictions in the report

        Returns
        -------
        str
            The formatted report to be print

        """
        return self._Report(self, keep_duplicates).to_string(show_subsets, show_questions)


def _get_name(content):
    return os.path.splitext(os.path.basename(content))[0]



# -*- coding: utf-8 -*-
"""Functions to evaluate some statistical properties of representations

"""

# ########################################
# Partition function
import numpy as np

import mangoes.utils.arrays


def _default_partition_function(embeddings, vector):
    """Default partition function

    Partition function used as default in function concentration_of_partition_function

    .. math::
        Z_c = \sum{\exp{c.w}}

    """
    return np.sum(np.exp(embeddings.matrix.dot(vector)))


def _generate_random_vectors(nb_vectors, dimension, mu):
    """Generate a list of uniformly random vectors
    """
    random_discourse_vectors = np.zeros(shape=(nb_vectors, dimension))
    for i in range(dimension):
        random_discourse_vectors[:, i] = np.random.normal(size=nb_vectors)
    return [(c / np.linalg.norm(c)) * (np.sqrt(dimension) / 5 / mu) for c in random_discourse_vectors]


def isotropy_from_partition_function(embeddings, discourse_vectors=1000,
                                     partition_function=_default_partition_function, epsilon=0.1):
    """Evaluate the isotropy of the representation computing the values of a partition function

    Compute the values of a repartition function (default: :math:`Z_c = \sum{\exp{cw}}`) for each :math:`c` in a set of
    uniformly random chosen vectors

    References
    ----------
    .. [1] Arora, S., Li, Y., Liang, Y., Ma, T., & Risteski, A. (2015). Rand-walk: A latent variable model approach to
           word embeddings.

    Parameters
    ----------
    embeddings: instance of Embeddings
        The :class:`mangoes.Embeddings` instance to evaluate
    discourse_vectors: int or list, optional
        a list of vectors or the number of vectors to pick uniformly on the sphere of norm 4/µ where µ is the average
        norm of the word vectors. Default : 1000.
    partition_function: callable
        the partition function to evaluate. Default : :func:`._default_partition_function`
    epsilon: float, optional
        determines the interval used to evaluate the concentration : [(1-ε)*mean_value, (1+ε)*mean_value]
        default: 0.1

    Returns
    -------
    tuple
        (concentration i.e. proportion of the values around the mean value, mean_value, values)

    """
    # TODO : only for dense matrices
    try:
        nb_random_vectors = len(discourse_vectors)
    except TypeError:
        nb_random_vectors = discourse_vectors
        discourse_vectors = _generate_random_vectors(nb_random_vectors,
                                                     embeddings.shape[1],
                                                     np.mean(np.linalg.norm(embeddings.matrix, axis=1)))

    partition_values = [partition_function(embeddings, c) for c in discourse_vectors]
    mean_value = np.mean(partition_values)
    nb_around_mean = (((1 - epsilon) * mean_value < partition_values)
                      & (partition_values < (1 + epsilon) * mean_value)).sum()
    return nb_around_mean / nb_random_vectors, mean_value, partition_values


# ########################################
# Distances between words

def _angles(vector, matrix, normalize=True):
    """Compute the angles between a vector and each line of the matrix

    Parameters
    ----------
    vector:
        a vector of dimension d
    matrix:
        a matrix n x d
    normalize: boolean, optional
        set to False if the matrix and the vectors are already normalized (default: True)

    Returns
    -------
    list
        list of angles size n
    """
    if normalize:
        matrix = matrix.normalize()
        vector = vector.normalize()

    cosines = matrix.dot(vector.T)
    try:
        cosines = cosines.todense()
    except AttributeError:
        pass

    # fix rounding errors
    cosines[[cosines > 1]] = 1
    cosines[[cosines < -1]] = -1

    return np.arccos(cosines)


def distances_one_word_histogram(embeddings, word, bins, distance=_angles, normalize=True):
    """Compute the values of an histogram of the distances between a word and all the other words of the Embeddings

    Parameters
    ----------

    embeddings: instance of Embeddings
        The :class:`mangoes.Embeddings` instance to evaluate
    word: str
        a word from the vocabulary of the Embedding
    bins: list
        bin edges, including the rightmost edge.
    distance: callable, optional
        the function to use to compute distances between words. Default: :func:`._angles`
    normalize: boolean, optional
        set to False if the matrix and the vectors are already normalized (default: True)

    Returns
    -------
    array
        an array of size (len(bins) - 1) with the values of the histogram
    """
    word_index = embeddings.words.word_index[word]
    matrix = mangoes.utils.arrays.Matrix.factory(
        np.concatenate((embeddings.matrix[:word_index], embeddings.matrix[word_index + 1:])))
    return np.histogram(distance(embeddings.matrix[word_index], matrix, normalize=normalize), bins=bins)[0]


def distances_histogram(embeddings, bins, distance=_angles, normalize=True):
    """Compute the values of an histogram of the distances between all the words of the Embeddings.

    Parameters
    ----------
    embeddings: instance of Embeddings
        The :class:`mangoes.Embeddings` instance to evaluate
    bins: list
        bin edges, including the rightmost edge.
    distance: callable
        the function to use to compute distances between words
    normalize: boolean
        set to False if the matrix and the vectors are already normalized (default: True)

    Returns
    -------
    array
        an array of size (len(bins) - 1) with the values of the histogram
    """
    if normalize:
        matrix = embeddings.matrix.normalize()
    else:
        matrix = embeddings.matrix

    hist = np.array([0] * (len(bins) - 1))
    for i in range(len(embeddings.words) - 1):
        vector = matrix[i]
        hist += np.histogram(distance(vector, matrix[i + 1:], normalize=False), bins=bins)[0]

    return hist
